/*****************************************************************************
 *                                                                           *
 *          UNURAN -- Universal Non-Uniform Random number generator          *
 *                                                                           *
 *****************************************************************************
 *                                                                           *
 *   FILE:      stringparser.c.in / stringparser.c                           *
 *                                                                           *
 *                                                                           *
 *   DESCRIPTION:                                                            *
 *                                                                           *
 *   The function unur_str2gen() takes a character string as its argument.   *
 *   This string is parsed in the following ways and the information         *
 *   obtained is used to make a generator object. (It returns NULL when      *
 *   it is not possible to make a generator object either due to a syntax    *
 *   error in the given string, or due to invalid data.                      *
 *                                                                           *
 *   First the given string is copied into a working string. Moreover letter *
 *   letters are transformed into their lower case counterparts and spaces   *
 *   are eliminated (i.e. the working string does not contain any white      *
 *   space characters any more.) Then the working string is splitted into    *
 *   several blocks, separated by ampersands '&'. Each block consists of     *
 *   <key>=<value> pairs, separated by semicolons ';'. The first key in      *
 *   each block is used to indicate the corresponding block.                 *
 *                                                                           *
 *   We have three different blocks (yet):                                   *
 *      distr  ... contains the definition of the distribution               *
 *      method ... contains the description of the transformation methdod    *
 *      urng   ... contains the definition of the uniform RNG                *
 *                                                                           *
 *   The 'distr' block must be the very first block and is obligatory.       *
 *   For that reason the keyword "distr" is optional and can be omitted      *
 *   (together with the '=' character). Moreover it is almost ignored while  *
 *   parsing the string. To avoid some possible confussion it only has to    *
 *   start with the letter 'd' (if it is given at all). E.g. "dnormal" can   *
 *   be used as keyword.                                                     *
 *   The value of the "distr" key is used to get a distribution object       *
 *   either via a unur_distr_<value>() call for a standard distribution      *
 *   or the unur_distr_<value>_new() calls are used to get an object for a   *
 *   generic distribution. However not all generic distributions are         *
 *   supported yet (see the list %UNSUPPORTED_DISTR_TYPES in file            *
 *   make_stringparser.pl).                                                  *
 *   The parameters for the standard distribution are given as a list of     *
 *   numbers enclosed in parenthesis. There must not be any character (other *
 *   than white space) between the name of the standard distribution and     *
 *   this list.                                                              *
 *                                                                           *
 *   All the other blocks are optional and can be arranged in arbitrary      *
 *   order.                                                                  *
 *   The value of the "method" key is the name of the transformation method  *
 *   and is used to execute the unur_<value>_new() call.                     *
 *                                                                           *
 *   The value of the "urng" key is passed to the PRNG interface.            *
 *   However it only works when using the PRNG library is enabled in         *
 *   unuran_config.h.                                                        *
 *                                                                           *
 *   In each block consecuting <key>=<value> pairs, separated by semicolons  *
 *   ';', are used to set parameters. The name of the parameter is given as  *
 *   key, the argument list is given as value. The arguments of the          *
 *   corresponding set calls are given as tokens, separated by commata ','.  *
 *   There are two types of tokens:                                          *
 *      single tokens, that represent numbers, and                           *
 *      list, i.e. a list of single tokens, separated by commata ',',        *
 *            enclosed in parenthesis.                                       *
 *   The value (including the character '=') can be omitted when no argument *
 *   are required.                                                           *
 *   The key is deduced from the UNURAN set calls, such that it uses only    *
 *   the part of the set call beyond "..._set_".                             *
 *   Not all set commands are supported yet (see the output to STDERR of     *
 *   the make_stringparser.pl script).                                       *
 *   As a rule of thumb only those set calls are supported that only have    *
 *   numbers or arrays of numbers as arguments (besides the pointer to the   *
 *   distribution or parameter object).                                      *
 *   (See the code of make_stringparser.pl for details.)                     *
 *                                                                           *
 *   There are additional transformations before executing the necessary     *
 *   set calls (listed by argument types)                                    *
 *      int:      'true' and 'on' are transformed to 1,                      *
 *                'false' and 'off' are transformed to 0.                    *
 *                a missing argument is transformed to 1.                    *
 *      int, int: Instead of two integers, a list with at least two          *
 *                numbers can be given.                                      *
 *                'inf[...]' is transformed to INT_MAX,                      *
 *                '-inf[...]' is transformed to INT_MIN.                     *
 *      unsigned: string is interpreted as hexadecimal number.               *
 *      double:   'inf[...]' is transformed to INFINITY,                     *
 *                '-inf[...]' is transformed to -INFINITY.                   *
 *      double, double: Instead of two doubles, a list with at least two     *
 *                numbers can be given.                                      *
 *                'inf[...]' is transformed to INFINITY.                     *
 *      int, double*: If the first argument is missing, it is replaced       *
 *                by the size of the array.                                  *
 *                If the second argument is missing, the NULL pointer is     *
 *                used instead an array as argument.                         *
 *      double*, int: Only for distributions!                                *
 *                If the second argument 'int' is missing, it is replaced    *
 *                by the size of the array.                                  *
 *   Important: there is no support for 'infinity' in lists!                 * 
 *   (See also the source below for more details.)                           *
 *                                                                           *
 *****************************************************************************
 *                                                                           *
 *   REMARK:                                                                 *
 *   This file stringpars.c.in contains the main part of the stringparser.   *
 *   The perl script make_stringparser.pl replaces the                       *
 *        =INPUT keyword                                                     *
 *   tags with search lists created from the information found within the    *
 *   header files of the source code.                                        *
 *   These lists (implemented via 'if' rules together with switch lists to   *
 *   with the first letter of the set calls as simple hash function) are     *
 *   used to call the corresponding ..._set and ..._new calls for the        *
 *   the keywords found in the string.                                       *
 *                                                                           *
 *****************************************************************************
 *                                                                           *
 *   REMARK:                                                                 *
 *   The stringparser always uses the default debugging flag.                *
 *                                                                           *
 *****************************************************************************
     $Id: stringparser.c.in,v 1.46 2004/07/14 17:50:58 leydold Exp $
 *****************************************************************************
 *                                                                           *
 *   Copyright (c) 2000 Wolfgang Hoermann and Josef Leydold                  *
 *   Dept. for Statistics, University of Economics, Vienna, Austria          *
 *                                                                           *
 *   This program is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the                           *
 *   Free Software Foundation, Inc.,                                         *
 *   59 Temple Place, Suite 330, Boston, MA 02111-1307, USA                  *
 *                                                                           *
 *****************************************************************************/

#include <stdarg.h>
#include <ctype.h>
#include <unur_source.h>
#include <unuran.h.in>
#include <distr/distr_source.h>
#include <methods/unur_methods_source.h>
#include "parser_source.h"

/*---------------------------------------------------------------------------*/
/* Variants: none                                                            */

/*---------------------------------------------------------------------------*/
/* Debugging flags                                                           */
/*    bit  01    ... pameters and structure of generator (do not use here)   */
/*    bits 02-12 ... setup                                                   */
/*    bits 13-24 ... adaptive steps                                          */
/*    bits 25-32 ... trace sampling                                          */

/*---------------------------------------------------------------------------*/
/* Flags for logging set calls                                               */

/*---------------------------------------------------------------------------*/

#define GENTYPE "STRING"       /* (pseudo) type of generator                 */

/*---------------------------------------------------------------------------*/

static struct unur_distr *_unur_str_distr( char *str_distr );
/*---------------------------------------------------------------------------*/
/* get distribution object for given distribution.                           */
/*---------------------------------------------------------------------------*/

static struct unur_distr *_unur_str_distr_new( char *distribution );
/*---------------------------------------------------------------------------*/
/* get new distribution object.                                              */
/*---------------------------------------------------------------------------*/

static struct unur_distr *_unur_str_distr_make_os( UNUR_DISTR *distr, 
						   const char *key,
						   char *type_args, char **args );
/*---------------------------------------------------------------------------*/
/* Make distribution object for order statistics of given distribution.      */
/*---------------------------------------------------------------------------*/

static int _unur_str_distr_set( UNUR_DISTR **ptr_distr, const char *key, char *value );
/*---------------------------------------------------------------------------*/
/* set parameters for distribution.                                          */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Set parameter in given distribution object.                               */
/*---------------------------------------------------------------------------*/
typedef int distr_set_i( UNUR_DISTR *distr, int i );
typedef int distr_set_ii( UNUR_DISTR *distr, int i1, int i2 );
typedef int distr_set_d( UNUR_DISTR *distr, double d );
typedef int distr_set_dd( UNUR_DISTR *distr, double d1, double d2 );
typedef int distr_set_Di( UNUR_DISTR *distr, const double *array, int size );
typedef int distr_set_C( UNUR_DISTR *distr, const char *string );
/*---------------------------------------------------------------------------*/
/* types of set calls                                                        */
/*---------------------------------------------------------------------------*/

static int _unur_str_distr_set_i( UNUR_DISTR *distr, const char *key, char *type_args, char **args,
				  distr_set_i set );
/*---------------------------------------------------------------------------*/
/* integer.                                                                  */
/*---------------------------------------------------------------------------*/

static int _unur_str_distr_set_ii( UNUR_DISTR *distr, const char *key, char *type_args, char **args,
				   distr_set_ii set );
/*---------------------------------------------------------------------------*/
/* 2 integers.                                                               */
/*---------------------------------------------------------------------------*/

static int _unur_str_distr_set_d( UNUR_DISTR *distr, const char *key, char *type_args, char **args,
				  distr_set_d set );
/*---------------------------------------------------------------------------*/
/* double.                                                                   */
/*---------------------------------------------------------------------------*/

static int _unur_str_distr_set_dd( UNUR_DISTR *distr, const char *key, char *type_args, char **args,
				   distr_set_dd set );
/*---------------------------------------------------------------------------*/
/* 2 doubles.                                                                */
/*---------------------------------------------------------------------------*/

static int _unur_str_distr_set_Di( UNUR_DISTR *distr, const char *key, char *type_args, char **args,
				   distr_set_Di set );
/*---------------------------------------------------------------------------*/
/* array of doubles & integer.                                               */
/*---------------------------------------------------------------------------*/

static int _unur_str_distr_set_C( UNUR_DISTR *distr, const char *key, char *type_args, char **args,
				  distr_set_C set );
/*---------------------------------------------------------------------------*/
/* constant string.                                                          */
/*---------------------------------------------------------------------------*/


static struct unur_par *_unur_str_par( char *str_method, const UNUR_DISTR *distr,
				       struct unur_slist *mlist );
/*---------------------------------------------------------------------------*/
/* get parameter object for distribution and method.                         */
/*---------------------------------------------------------------------------*/

static struct unur_par *_unur_str_par_new( const char *method, const UNUR_DISTR *distr );
/*---------------------------------------------------------------------------*/
/* get new parameter object for method.                                      */
/*---------------------------------------------------------------------------*/

static int _unur_str_par_set( UNUR_PAR *par, const char *key, char *value,
			      struct unur_slist *mlist );
/*---------------------------------------------------------------------------*/
/* set parameters for method.                                                */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Set parameter in given parameter object.                                 */
/*---------------------------------------------------------------------------*/
typedef int par_set_void( UNUR_PAR *par );
typedef int par_set_i( UNUR_PAR *par, int i );
typedef int par_set_u( UNUR_PAR *par, unsigned u );
typedef int par_set_d( UNUR_PAR *par, double d );
typedef int par_set_dd( UNUR_PAR *par, double d1, double d2 );
typedef int par_set_iD( UNUR_PAR *par, int size, const double *array );
typedef int par_set_Di( UNUR_PAR *par, const double *array, int size );
/*---------------------------------------------------------------------------*/
/* types of set calls                                                        */
/*---------------------------------------------------------------------------*/

static int _unur_str_par_set_void( UNUR_PAR *par, const char *key, char *type_args, char **args,
				   par_set_void set );
/*---------------------------------------------------------------------------*/
/* void (no argument required).                                              */
/*---------------------------------------------------------------------------*/

static int _unur_str_par_set_i( UNUR_PAR *par, const char *key, char *type_args, char **args,
				par_set_i set );
/*---------------------------------------------------------------------------*/
/* integer.                                                                  */
/*---------------------------------------------------------------------------*/

static int _unur_str_par_set_u( UNUR_PAR *par, const char *key, char *type_args, char **args,
				par_set_u set );
/*---------------------------------------------------------------------------*/
/* unsigned integer.                                                         */
/*---------------------------------------------------------------------------*/

static int _unur_str_par_set_d( UNUR_PAR *par, const char *key, char *type_args, char **args,
				par_set_d set );
/*---------------------------------------------------------------------------*/
/* double.                                                                   */
/*---------------------------------------------------------------------------*/

static int _unur_str_par_set_dd( UNUR_PAR *par, const char *key, char *type_args, char **args,
				 par_set_dd set );

/*---------------------------------------------------------------------------*/
/* 2 doubles.                                                                */
/*---------------------------------------------------------------------------*/

static int _unur_str_par_set_iD( UNUR_PAR *par, const char *key, char *type_args, char **args,
				 par_set_iD set, struct unur_slist *mlist );
/*---------------------------------------------------------------------------*/
/* integer & array of doubles.                                               */
/*---------------------------------------------------------------------------*/

static int _unur_str_par_set_Di( UNUR_PAR *par, const char *key, char *type_args, char **args,
				 par_set_Di set, struct unur_slist *mlist );
/*---------------------------------------------------------------------------*/
/* array of doubles & integer.                                               */
/*---------------------------------------------------------------------------*/


static UNUR_URNG *_unur_str2urng( char *str_urng );
/*---------------------------------------------------------------------------*/
/* get uniform RNG.                                                          */
/*---------------------------------------------------------------------------*/


static int _unur_str_set_args( char *value, char *type_args, char **args, int max_args );
/*---------------------------------------------------------------------------*/
/* parse argument string for set call.                                       */
/*---------------------------------------------------------------------------*/

static int _unur_parse_ilist( char *liststr, int **iarray );
/*---------------------------------------------------------------------------*/
/* Process a comma separated list of integers.                               */
/*---------------------------------------------------------------------------*/

static int _unur_parse_dlist( char *liststr, double **darray );
/*---------------------------------------------------------------------------*/
/* Process a comma separated list of numbers.                                */
/*---------------------------------------------------------------------------*/

static int _unur_atoi ( const char *str );
/*---------------------------------------------------------------------------*/
/* Convert string into its integer representation.                           */
/*---------------------------------------------------------------------------*/

static unsigned _unur_atou ( const char *str );
/*---------------------------------------------------------------------------*/
/* Convert string into its unsigned representation (using hexadecimal).      */
/*---------------------------------------------------------------------------*/

static double _unur_atod ( const char *str );
/*---------------------------------------------------------------------------*/
/* Convert string into its double value.                                     */
/*---------------------------------------------------------------------------*/

#ifdef UNUR_ENABLE_LOGGING
/*---------------------------------------------------------------------------*/
/* the following functions print debugging information on output stream,     */
/* i.e., into the log file if not specified otherwise.                       */
/*---------------------------------------------------------------------------*/

static void _unur_str_debug_string( int level, const char *key, const char *value );
/*---------------------------------------------------------------------------*/
/* print key & value info into logfile.                                      */
/*---------------------------------------------------------------------------*/

static void _unur_str_debug_distr( int level, const char *name, double *params, int n_params );
/*---------------------------------------------------------------------------*/
/* write info about distribution into logfile.                               */
/*---------------------------------------------------------------------------*/

static void _unur_str_debug_set( int level, const char *key, const char *type, ... );
/*---------------------------------------------------------------------------*/
/* write info about set command into logfile.                                */
/*---------------------------------------------------------------------------*/

#endif

/*---------------------------------------------------------------------------*/
/* abbreviations */

/* unknown keyword */
#define _unur_error_unknown(key,what) \
   do { \
     unur_errno = UNUR_ERR_STR_UNKNOWN; \
     _unur_stream_printf ( GENTYPE,__FILE__,__LINE__, \
			   "error: %s : unknown %s: %s", \
			   unur_get_strerror(unur_errno), (what), (key) ); \
   } while (0)

/* invalid data for key */
#define _unur_error_invalid(key,what) \
   do { \
     unur_errno = UNUR_ERR_STR_INVALID; \
     _unur_stream_printf ( GENTYPE,__FILE__,__LINE__, \
			   "error: %s: invalid data for %s %s", \
			   unur_get_strerror(unur_errno), (what), (key) ); \
   } while (0)

/* invalid argument string for set calls */
#define _unur_error_args(key) \
   do { \
     unur_errno = UNUR_ERR_STR_INVALID; \
     _unur_stream_printf ( GENTYPE,__FILE__,__LINE__, \
			   "error: %s: invalid argument string for %s", \
			   unur_get_strerror(unur_errno), (key) ); \
   } while (0)


/*---------------------------------------------------------------------------*/
/* constants                                                                 */

#define MAX_SET_ARGS  (10)  
/* maximal number of arguments for set calls (must be at least 2) */

/*---------------------------------------------------------------------------*/

/*****************************************************************************/
/**  User Interface                                                         **/
/*****************************************************************************/

/*---------------------------------------------------------------------------*/

struct unur_gen *
unur_str2gen (const char *string)
     /*----------------------------------------------------------------------*/
     /* get generator object for distribution and method                     */
     /*                                                                      */
     /* parameters:                                                          */
     /*   string ... string that contains description of generator           */
     /*                                                                      */
     /* return:                                                              */
     /*   generator object                                                   */
     /*                                                                      */
     /* error:                                                               */
     /*   return NULL                                                        */
     /*----------------------------------------------------------------------*/
{
  UNUR_DISTR *distr = NULL;       /* distribution object */
  UNUR_PAR   *par   = NULL;       /* parameter object */
  UNUR_GEN   *gen   = NULL;       /* generator object    */
  UNUR_URNG  *urng  = NULL;       /* uniform random number generator */

  char *str_distr   = NULL;       /* string for distribution */
  char *str_method  = NULL;       /* string for method       */
  char *str_urng    = NULL;       /* string for uniform RNG  */

  char *str = NULL;               /* pointer to working string */
  char *token;

  struct unur_slist *mlist;       /* list of allocated memory blocks */

  /* check arguments */
  _unur_check_NULL( GENTYPE,string,NULL );

  /* (empty) list of allocated memory blocks */ 
  mlist = _unur_slist_new();

  /* Prepare string for processing:            */
  /*   Remove all white spaces,                */
  /*   convert to lower case letters,          */
  /*   copy into working string                */
  str = _unur_parser_prepare_string( string );

#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag)
    _unur_str_debug_string(0,"[input]",str);
#endif

  /* split string into blocks separated by ampersands '&' */

  /* the first block must contain the distribution */
  str_distr = strtok(str, "&");

  /* new get all the other tokens */
  for ( token  = strtok(NULL, "&"); 
        token != NULL;     
        token  = strtok(NULL, "&") ) {
    
    /* get type of block */
    if (!strncmp(token,"method=",7)) {
      str_method = token;
    }
    else if (!strncmp(token,"urng=",5)) {
      str_urng = token;
    }
    else {
      _unur_error_unknown(token,"category");
      _unur_slist_free( mlist );
      if (str) free(str);
      return NULL;
    }
  }
  
  /* make distribution object */
  distr = _unur_str_distr(str_distr);
  if ( distr == NULL ) {
    /* error */
    _unur_slist_free( mlist );
    if (str) free(str);
    return NULL;
  }

  /* get parameter object */
  if ( str_method != NULL )
    /* method is provided in string */
    par = _unur_str_par(str_method, distr, mlist);
  else
    /* otherwise use default method */
    par = unur_auto_new(distr);

  /* make generator object */
  gen = unur_init(par);

  /* destroy distribution object */
  unur_distr_free(distr);

  /* set uniform random number generator -- if provided */
  if ( str_urng != NULL )
    if (gen != NULL) {
      /* we need a valid generator object */
      if ((urng = _unur_str2urng(str_urng)) != NULL )
	unur_chg_urng(gen, urng);
    }

  /* free allocated memory blocks */
  _unur_slist_free(mlist);
  if (str) free(str);

  /* check result: nothing to do */
  /*  if ( gen == NULL ) { */
  /*    error */
  /*    return NULL; */
  /*  } */

#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag)
    _unur_str_debug_string(0,"",NULL);
#endif

  /* return pointer to generator object */
  return gen;

} /* end of unur_str2gen() */

/*---------------------------------------------------------------------------*/

struct unur_par *
_unur_str2par (const struct unur_distr *distr, const char *string, struct unur_slist **mlist )
     /*----------------------------------------------------------------------*/
     /* get parameter object for distribution and method                     */
     /*                                                                      */
     /* parameters:                                                          */
     /*   distr  ... pointer to distribution object                          */
     /*   string ... string that contains description of method              */
     /*   mlist  ... pointer to simple list for allocated memory blocks      */
     /*                                                                      */
     /* return:                                                              */
     /*   parameter object                                                   */
     /*                                                                      */
     /* error:                                                               */
     /*   return NULL                                                        */
     /*----------------------------------------------------------------------*/
{
  UNUR_PAR *par = NULL;           /* parameter object */
  char *str = NULL;               /* pointer to working string */

  /* check arguments */
  _unur_check_NULL( GENTYPE,distr,NULL );
  _unur_check_NULL( GENTYPE,string,NULL );

  /* (empty) list of allocated memory blocks */ 
  *mlist = _unur_slist_new();

  /* Prepare string for processing:            */
  /*   Remove all white spaces,                */
  /*   convert to lower case letters,          */
  /*   copy into working string                */
  str = _unur_parser_prepare_string( string );

#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag)
    _unur_str_debug_string(0,"[input]",str);
#endif

  /* get parameter object */
  par = _unur_str_par(str, distr, *mlist);

  /* free allocated memory blocks */
  if (str) free(str);

#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag)
    if ( par != NULL )
      _unur_str_debug_string(0,"",NULL);
#endif

  /* return pointer to parameter object */
  return par;

} /* end of unur_str2par() */

/*---------------------------------------------------------------------------*/

struct unur_distr *
unur_str2distr (const char *string)
     /*----------------------------------------------------------------------*/
     /* get distribution object for given distribution                       */
     /*                                                                      */
     /* parameters:                                                          */
     /*   string ... string that contains description of distribution        */
     /*                                                                      */
     /* return:                                                              */
     /*   distribution object                                                */
     /*                                                                      */
     /* error:                                                               */
     /*   return NULL                                                        */
     /*----------------------------------------------------------------------*/
{
  UNUR_DISTR *distr = NULL;       /* distribution object */
  char *str = NULL;               /* pointer to working string */

  /* check arguments */
  _unur_check_NULL( GENTYPE,string,NULL );

  /* Prepare string for processing:            */
  /*   Remove all white spaces,                */
  /*   convert to lower case letters,          */
  /*   copy into working string                */
  str = _unur_parser_prepare_string( string );

#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag)
    _unur_str_debug_string(0,"[input]",str);
#endif

  /* make distribution object */
  distr = _unur_str_distr(str);

  /* free allocated memory blocks */
  if (str) free(str);

#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag)
    if ( distr != NULL )
      _unur_str_debug_string(0,"",NULL);
#endif

  /* return pointer to distribution object */
  return distr;

} /* end of unur_str2distr() */

/*---------------------------------------------------------------------------*/

/*****************************************************************************/
/**  Distributions                                                          **/
/*****************************************************************************/

/*---------------------------------------------------------------------------*/

struct unur_distr *
_unur_str_distr( char *str_distr )
     /*----------------------------------------------------------------------*/
     /* get distribution object for given distribution                       */
     /*                                                                      */
     /* parameters:                                                          */
     /*   str_distr ... string that contains description of distribution     */
     /*                                                                      */
     /* return:                                                              */
     /*   distribution object                                                */
     /*                                                                      */
     /* error:                                                               */
     /*   return NULL                                                        */
     /*----------------------------------------------------------------------*/
{
  struct unur_distr *distr = NULL;

  char *token;             /* pointer to token in string */
  char *next;              /* pointer to next token in string */
  char *key, *value;       /* the key and its value */


  /* tokenize the string using ';' as separator and handle the tokens */
  for ( token = next = str_distr;
	next != NULL && *token != '\0';
	token = next ) {

    /* find end of token (i.e. the pointer to next token) */
    next = strchr(token,';');
    if (next != NULL) {
      /* next separator found */
      *next = '\0';     /* terminate token string    */
      next++;           /* set pointer to next token */
    }

    /* token points to a key */
    key = token;

    /* get the value from the pair key=value and terminate key string */
    value = strchr(key, '=');
    if (value != NULL) {
      *value = '\0';    /* terminate key string        */
      value++;          /* set pointer to value string */
    }

    /* the first key is treated differently: get new distribution object */
    if (key == str_distr) {
      /* the keyword "distr" is optional */
      if (value == NULL) {
	/* no "distribution" keyword given, so the key points to the value */
	value = key;
      }
      else {
	if (*key != 'd') {
	  _unur_error(GENTYPE,UNUR_ERR_STR_SYNTAX,"key for distribution does not start with 'd'"); 
	  return NULL;
	}
      }

      /* get new distribution object */
      distr = _unur_str_distr_new(value);
      if (distr == NULL) {
	/* error */
	/* _unur_error(GENTYPE,UNUR_ERR_STR,"setting distribution failed");  */
	return NULL;
      }
    }

    /* set parameters for distribution */
    else {
      if (_unur_str_distr_set(&distr, key, value)!=UNUR_SUCCESS ) {
	/* error */
	_unur_distr_free(distr);
	/* _unur_error(GENTYPE,UNUR_ERR_STR,"setting distribution failed");  */
	return NULL;
      }
    }
  }

  return distr;

} /* end of _unur_str_distr() */

/*---------------------------------------------------------------------------*/

struct unur_distr *
_unur_str_distr_new( char *distribution )
     /*----------------------------------------------------------------------*/
     /* get new distribution object                                          */
     /*                                                                      */
     /* parameters:                                                          */
     /*   distribution ... string that contains distribution name            */
     /*   params       ... string that contains list of parameters           */
     /*                                                                      */
     /* return:                                                              */
     /*   distribution object (pointer to structure)                         */
     /*                                                                      */
     /* error:                                                               */
     /*   return NULL                                                        */
     /*----------------------------------------------------------------------*/
{
  struct unur_distr *distr = NULL;    /* pointer to distribution object      */
  char distr_unknown;

  char *name;                         /* pointer to name of distribution     */
  char *params;                       /* pointer to parameter list of distr  */

  double *darray = NULL;              /* array of arguments for distribution */
  int n_darray = 0;                   /* size of array of parameters         */

  /* name of distribution */
  name = distribution;

  /* get parameter list */
  params = strchr(distribution,'(');
  if (params != NULL) {
    *params = '\0';                   /* terminate key string        */
    ++params;                         /* set pointer to value string */
  }

  /* get parameter list */
  n_darray = _unur_parse_dlist(params, &darray );

#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
    _unur_str_debug_distr(1,name,darray,n_darray);
#endif

  /* mark distribution as unknown (this is a very ugly hack) */
  distr = (struct unur_distr *) &distr_unknown;

	 switch (*distribution) {
	 case 'b':
		 if ( !strcmp( distribution, "beta") ) {
			 distr = unur_distr_beta (darray,n_darray);
			 break;
		 }
		 if ( !strcmp( distribution, "binomial") ) {
			 distr = unur_distr_binomial (darray,n_darray);
			 break;
		 }
		 break;
	 case 'c':
		 if ( !strcmp( distribution, "cauchy") ) {
			 distr = unur_distr_cauchy (darray,n_darray);
			 break;
		 }
		 if ( !strcmp( distribution, "chi") ) {
			 distr = unur_distr_chi (darray,n_darray);
			 break;
		 }
		 if ( !strcmp( distribution, "chisquare") ) {
			 distr = unur_distr_chisquare (darray,n_darray);
			 break;
		 }
		 break;
	 case 'e':
		 if ( !strcmp( distribution, "exponential") ) {
			 distr = unur_distr_exponential (darray,n_darray);
			 break;
		 }
		 if ( !strcmp( distribution, "extremei") ) {
			 distr = unur_distr_extremeI (darray,n_darray);
			 break;
		 }
		 if ( !strcmp( distribution, "extremeii") ) {
			 distr = unur_distr_extremeII (darray,n_darray);
			 break;
		 }
		 break;
	 case 'g':
		 if ( !strcmp( distribution, "gamma") ) {
			 distr = unur_distr_gamma (darray,n_darray);
			 break;
		 }
		 if ( !strcmp( distribution, "geometric") ) {
			 distr = unur_distr_geometric (darray,n_darray);
			 break;
		 }
		 break;
	 case 'h':
		 if ( !strcmp( distribution, "hypergeometric") ) {
			 distr = unur_distr_hypergeometric (darray,n_darray);
			 break;
		 }
		 break;
	 case 'l':
		 if ( !strcmp( distribution, "laplace") ) {
			 distr = unur_distr_laplace (darray,n_darray);
			 break;
		 }
		 if ( !strcmp( distribution, "logarithmic") ) {
			 distr = unur_distr_logarithmic (darray,n_darray);
			 break;
		 }
		 if ( !strcmp( distribution, "logistic") ) {
			 distr = unur_distr_logistic (darray,n_darray);
			 break;
		 }
		 if ( !strcmp( distribution, "lomax") ) {
			 distr = unur_distr_lomax (darray,n_darray);
			 break;
		 }
		 break;
	 case 'n':
		 if ( !strcmp( distribution, "negativebinomial") ) {
			 distr = unur_distr_negativebinomial (darray,n_darray);
			 break;
		 }
		 if ( !strcmp( distribution, "normal") ) {
			 distr = unur_distr_normal (darray,n_darray);
			 break;
		 }
		 break;
	 case 'p':
		 if ( !strcmp( distribution, "pareto") ) {
			 distr = unur_distr_pareto (darray,n_darray);
			 break;
		 }
		 if ( !strcmp( distribution, "poisson") ) {
			 distr = unur_distr_poisson (darray,n_darray);
			 break;
		 }
		 if ( !strcmp( distribution, "powerexponential") ) {
			 distr = unur_distr_powerexponential (darray,n_darray);
			 break;
		 }
		 break;
	 case 'r':
		 if ( !strcmp( distribution, "rayleigh") ) {
			 distr = unur_distr_rayleigh (darray,n_darray);
			 break;
		 }
		 break;
	 case 's':
		 if ( !strcmp( distribution, "student") ) {
			 distr = unur_distr_student (darray,n_darray);
			 break;
		 }
		 break;
	 case 't':
		 if ( !strcmp( distribution, "triangular") ) {
			 distr = unur_distr_triangular (darray,n_darray);
			 break;
		 }
		 break;
	 case 'u':
		 if ( !strcmp( distribution, "uniform") ) {
			 distr = unur_distr_uniform (darray,n_darray);
			 break;
		 }
		 break;
	 case 'w':
		 if ( !strcmp( distribution, "weibull") ) {
			 distr = unur_distr_weibull (darray,n_darray);
			 break;
		 }
	 }

	 /* get pointer to generic distribution object */
	 if (distr == (struct unur_distr *) &distr_unknown) { 
		 do {
			 if ( !strcmp( distribution, "cemp") ) {
				 distr = unur_distr_cemp_new();
				 break;
			 }
			 if ( !strcmp( distribution, "cont") ) {
				 distr = unur_distr_cont_new();
				 break;
			 }
			 if ( !strcmp( distribution, "discr") ) {
				 distr = unur_distr_discr_new();
				 break;
			 }
		 } while (0);
	 }


   if (distr == (struct unur_distr *) &distr_unknown) {
     /* unknown method */
     _unur_error_unknown(distribution,"distribution");
     distr = NULL;
   }
   else if (distr == NULL) {
     /* invalid data for chosen method */
     _unur_error_invalid(distribution,"distribution");
   }

   /* clear memory */
   if (darray) free(darray);

   /* return result */
   return distr;

} /* end of _unur_str_distr_new() */

/*---------------------------------------------------------------------------*/

static struct unur_distr *
_unur_str_distr_make_os (UNUR_DISTR *distr, const char *key, char *type_args, char **args)
     /*----------------------------------------------------------------------*/
     /* Make distribution object for order statistics of given distribution. */
     /* The given distribution `distr' is destroyed.                         */
     /* The list of arguments and their types are given as string.           */
     /*                                                                      */
     /* type: "ii"   (int,int)                                               */
     /*                                                                      */
     /* input: "tt"  (exactly two single tokens)                             */
     /*        "L"   (exactly one list)                                      */
     /*                                                                      */
     /* action:                                                              */
     /*    if "tt": convert both to int.                                     */
     /*             execute set call.                                        */
     /*    if "L":  convert list entries to int.                             */
     /*             execute set call with first two entries.                 */
     /*                (error if list has less than two entries)             */
     /*    otherwise: error                                                  */
     /*                                                                      */
     /* parameters:                                                          */
     /*   distr     ... pointer to distribution object                       */
     /*   key       ... name of parameter to be set                          */
     /*   type_args ... contains types of arguments. Possible values:        */
     /*                   't'   for single token                             */
     /*                   'L'   for a list                                   */
     /*   args      ... array of pointers to argument strings                */
     /*                                                                      */
     /* return:                                                              */
     /*   pointer to distribution object for order statistics                */
     /*                                                                      */
     /* error:                                                               */
     /*   return NULL                                                        */
     /*----------------------------------------------------------------------*/
{
  int *iarray = NULL;                 /* pointer to working array            */
  struct unur_distr *os = NULL;       /* pointer to order statistics object  */

  /* either exactly two arguments are given */
  if ( !strcmp(type_args, "tt") ) {
    iarray = _unur_xmalloc( 2*sizeof(double) );
    iarray[0] = _unur_atoi( args[0] );
    iarray[1] = _unur_atoi( args[1] );
  }

  /* or one list with 2 entries is given */
  else if ( !strcmp(type_args, "L") ) {
    if ( _unur_parse_ilist( args[0], &iarray ) < 2 ) {
      free (iarray);
      iarray = NULL;
    }
  }

  if (iarray == NULL ) {
    /* data list empty (incorrect syntax) */
    _unur_error_args(key);
#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"!");
#endif
    return NULL;
  }

  /* now make object to order statistics */

#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
    _unur_str_debug_set(2,key,"ii",iarray[0],iarray[1] );
#endif

  os =  unur_distr_corder_new( distr, iarray[0], iarray[1] );

  /* free memory */
  _unur_distr_free(distr);
  free (iarray);
  
  /* return order statistics (or NULL we could not make such a thing) */
  return os;

} /* end of _unur_str_distr_make_os() */

/*---------------------------------------------------------------------------*/

int
_unur_str_distr_set( UNUR_DISTR **ptr_distr, const char *key, char *value )
     /*----------------------------------------------------------------------*/
     /* set parameters for distribution                                      */
     /*                                                                      */
     /* it also makes a distribution object for an order statistics for      */
     /* the given distribution when the key word "orderstatistics" occurs.   */
     /* Thus the distribution object itself might be changed.                */
     /*                                                                      */
     /* parameters:                                                          */
     /*   ptr_distr ... holds pointer to distribution object                 */
     /*   key       ... string that contains key for parameter               */
     /*   value     ... string that contains list of arguments for parameter */
     /*                                                                      */
     /* return:                                                              */
     /*   UNUR_SUCCESS ... on success                                        */
     /*   error code   ... on error                                          */
     /*                                                                      */
     /* error:                                                               */
     /*   return error code                                                  */
     /*----------------------------------------------------------------------*/
{
  int result;      /* result of UNURAN set call */

  /* derefence pointer to distribution object */
  struct unur_distr *distr = *ptr_distr;

  /* storing arguments of set calls: */
  char type_args[MAX_SET_ARGS+1];  /* array containing info about arguments */
  char *args[MAX_SET_ARGS+1];      /* pointers to argument strings */

  /* tokenize argument string */
  if (_unur_str_set_args( value, type_args, args, MAX_SET_ARGS ) < 0) {
    /* error */
    return UNUR_ERR_STR_SYNTAX;
  }

  /* set result indicator to unknown */
  result = UNUR_ERR_STR_UNKNOWN;

  /* find and execute set call */

	 switch (distr->type) {
	 case UNUR_DISTR_CEMP:
		 switch (*key) {
		 case 'd':
			 if ( !strcmp(key, "data") ) {
				 /* n = 2; type = Di:  UNUR_DISTR *distribution, const double *sample, int n_sample */
				 result = _unur_str_distr_set_Di(distr,key,type_args,args,unur_distr_cemp_set_data);
				 break;
			 }
		 }
		 break;
	 case UNUR_DISTR_CONT:
		 switch (*key) {
		 case 'c':
			 if ( !strcmp(key, "cdf") ) {
				 /* n = 1; type = C:  UNUR_DISTR *distribution, const char *cdfstr */
				 result = _unur_str_distr_set_C(distr,key,type_args,args,unur_distr_cont_set_cdfstr);
				 break;
			 }
			 if ( !strcmp(key, "cdfstr") ) {
				 /* n = 1; type = C:  UNUR_DISTR *distribution, const char *cdfstr */
				 result = _unur_str_distr_set_C(distr,key,type_args,args,unur_distr_cont_set_cdfstr);
				 break;
			 }
			 break;
		 case 'd':
			 if ( !strcmp(key, "domain") ) {
				 /* n = 2; type = dd:  UNUR_DISTR *distribution, double left, double right */
				 result = _unur_str_distr_set_dd(distr,key,type_args,args,unur_distr_cont_set_domain);
				 break;
			 }
			 break;
		 case 'h':
			 if ( !strcmp(key, "hr") ) {
				 /* n = 1; type = C:  UNUR_DISTR *distribution, const char *hrstr */
				 result = _unur_str_distr_set_C(distr,key,type_args,args,unur_distr_cont_set_hrstr);
				 break;
			 }
			 if ( !strcmp(key, "hrstr") ) {
				 /* n = 1; type = C:  UNUR_DISTR *distribution, const char *hrstr */
				 result = _unur_str_distr_set_C(distr,key,type_args,args,unur_distr_cont_set_hrstr);
				 break;
			 }
			 break;
		 case 'm':
			 if ( !strcmp(key, "mode") ) {
				 /* n = 1; type = d:  UNUR_DISTR *distribution, double mode */
				 result = _unur_str_distr_set_d(distr,key,type_args,args,unur_distr_cont_set_mode);
				 break;
			 }
			 break;
		 case 'p':
			 if ( !strcmp(key, "pdf") ) {
				 /* n = 1; type = C:  UNUR_DISTR *distribution, const char *pdfstr */
				 result = _unur_str_distr_set_C(distr,key,type_args,args,unur_distr_cont_set_pdfstr);
				 break;
			 }
			 if ( !strcmp(key, "pdfarea") ) {
				 /* n = 1; type = d:  UNUR_DISTR *distribution, double area */
				 result = _unur_str_distr_set_d(distr,key,type_args,args,unur_distr_cont_set_pdfarea);
				 break;
			 }
			 if ( !strcmp(key, "pdfparams") ) {
				 /* n = 2; type = Di:  UNUR_DISTR *distribution, const double *params, int n_params */
				 result = _unur_str_distr_set_Di(distr,key,type_args,args,unur_distr_cont_set_pdfparams);
				 break;
			 }
			 if ( !strcmp(key, "pdfstr") ) {
				 /* n = 1; type = C:  UNUR_DISTR *distribution, const char *pdfstr */
				 result = _unur_str_distr_set_C(distr,key,type_args,args,unur_distr_cont_set_pdfstr);
				 break;
			 }
		 }
		 break;
	 case UNUR_DISTR_DISCR:
		 switch (*key) {
		 case 'c':
			 if ( !strcmp(key, "cdf") ) {
				 /* n = 1; type = C:  UNUR_DISTR *distribution, const char *cdfstr */
				 result = _unur_str_distr_set_C(distr,key,type_args,args,unur_distr_discr_set_cdfstr);
				 break;
			 }
			 if ( !strcmp(key, "cdfstr") ) {
				 /* n = 1; type = C:  UNUR_DISTR *distribution, const char *cdfstr */
				 result = _unur_str_distr_set_C(distr,key,type_args,args,unur_distr_discr_set_cdfstr);
				 break;
			 }
			 break;
		 case 'd':
			 if ( !strcmp(key, "domain") ) {
				 /* n = 2; type = ii:  UNUR_DISTR *distribution, int left, int right */
				 result = _unur_str_distr_set_ii(distr,key,type_args,args,unur_distr_discr_set_domain);
				 break;
			 }
			 break;
		 case 'm':
			 if ( !strcmp(key, "mode") ) {
				 /* n = 1; type = i:  UNUR_DISTR *distribution, int mode */
				 result = _unur_str_distr_set_i(distr,key,type_args,args,unur_distr_discr_set_mode);
				 break;
			 }
			 break;
		 case 'p':
			 if ( !strcmp(key, "pmf") ) {
				 /* n = 1; type = C:  UNUR_DISTR *distribution, const char *pmfstr */
				 result = _unur_str_distr_set_C(distr,key,type_args,args,unur_distr_discr_set_pmfstr);
				 break;
			 }
			 if ( !strcmp(key, "pmfparams") ) {
				 /* n = 2; type = Di:  UNUR_DISTR *distribution, const double *params, int n_params */
				 result = _unur_str_distr_set_Di(distr,key,type_args,args,unur_distr_discr_set_pmfparams);
				 break;
			 }
			 if ( !strcmp(key, "pmfstr") ) {
				 /* n = 1; type = C:  UNUR_DISTR *distribution, const char *pmfstr */
				 result = _unur_str_distr_set_C(distr,key,type_args,args,unur_distr_discr_set_pmfstr);
				 break;
			 }
			 if ( !strcmp(key, "pmfsum") ) {
				 /* n = 1; type = d:  UNUR_DISTR *distribution, double sum */
				 result = _unur_str_distr_set_d(distr,key,type_args,args,unur_distr_discr_set_pmfsum);
				 break;
			 }
			 if ( !strcmp(key, "pv") ) {
				 /* n = 2; type = Di:  UNUR_DISTR *distribution, const double *pv, int n_pv */
				 result = _unur_str_distr_set_Di(distr,key,type_args,args,unur_distr_discr_set_pv);
				 break;
			 }
		 }
		 break;
	 }

	 /* set calls for all distribution types */
	 if (result == UNUR_ERR_STR_UNKNOWN) {
		 switch (*key) {
		 case 'n':
			 if ( !strcmp(key, "name") ) {
				 /* n = 1; type = C:  UNUR_DISTR *distribution, const char *name */
				 result = _unur_str_distr_set_C(distr,key,type_args,args,unur_distr_set_name);
				 break;
			 }
		 }
	 }


	 /* Ignored set commands: */
 	 /* int unur_distr_cont_set_pdf( UNUR_DISTR *distribution, UNUR_FUNCT_CONT *pdf );
		 n = 1; type = ?	 */
	 /* int unur_distr_cont_set_dpdf( UNUR_DISTR *distribution, UNUR_FUNCT_CONT *dpdf );
		 n = 1; type = ?	 */
	 /* int unur_distr_cont_set_cdf( UNUR_DISTR *distribution, UNUR_FUNCT_CONT *cdf );
		 n = 1; type = ?	 */
	 /* int unur_distr_cont_set_hr( UNUR_DISTR *distribution, UNUR_FUNCT_CONT *hazard );
		 n = 1; type = ?	 */
	 /* int unur_distr_discr_set_pmf( UNUR_DISTR *distribution, UNUR_FUNCT_DISCR *pmf );
		 n = 1; type = ?	 */
	 /* int unur_distr_discr_set_cdf( UNUR_DISTR *distribution, UNUR_FUNCT_DISCR *cdf );
		 n = 1; type = ?	 */


	 /* Subsituted set commands: */
 	 /* int unur_distr_cont_set_pdfstr( UNUR_DISTR *distribution, const char *pdfstr );
		 n = 1; type = C	 */
	 /* int unur_distr_cont_set_cdfstr( UNUR_DISTR *distribution, const char *cdfstr );
		 n = 1; type = C	 */
	 /* int unur_distr_cont_set_hrstr( UNUR_DISTR *distribution, const char *hrstr );
		 n = 1; type = C	 */
	 /* int unur_distr_discr_set_pmfstr( UNUR_DISTR *distribution, const char *pmfstr );
		 n = 1; type = C	 */
	 /* int unur_distr_discr_set_cdfstr( UNUR_DISTR *distribution, const char *cdfstr );
		 n = 1; type = C	 */


   /* special keyword */
   if (result == UNUR_ERR_STR_UNKNOWN)
     if (distr->type == UNUR_DISTR_CONT) {
       if ( !strcmp(key, "orderstatistics") ) {
	 /* make order statistics and replace distribution object */
	 *ptr_distr = _unur_str_distr_make_os (distr, key, type_args, args);
	 result = (*ptr_distr == NULL) ? UNUR_ERR_STR_SYNTAX : UNUR_SUCCESS;
       }
     }
  
  if (result == UNUR_ERR_STR_UNKNOWN) {
    /* unknown parameter */
    _unur_error_unknown(key,"parameter for given distribution");
    return UNUR_ERR_STR_UNKNOWN;
  }
  else if (result != UNUR_SUCCESS) {
    /* invalid data for parameter */
    _unur_error_invalid(key,"set call");
    return UNUR_ERR_STR_SYNTAX;
  }
  
  return UNUR_SUCCESS;
  
} /* end of _unur_str_distr_set() */

/*---------------------------------------------------------------------------*/

int
_unur_str_distr_set_i (UNUR_DISTR *distr, const char *key, char *type_args, char **args, distr_set_i set)
     /*----------------------------------------------------------------------*/
     /* Set parameter in given distribution object.                          */
     /* The list of arguments and their types are given as string.           */
     /*                                                                      */
     /* type: "i"   (int)                                                    */
     /*                                                                      */
     /* input: "t"  (exactly one single token)                               */
     /*        void (no argument given)                                      */
     /*                                                                      */
     /* action:                                                              */
     /*    convert token to int.                                             */
     /*      (if no argument is given, use 1.)                               */
     /*    execute set call.                                                 */
     /*    error if input != "t"                                             */
     /*                                                                      */
     /* parameters:                                                          */
     /*   distr     ... pointer to distribution object                       */
     /*   key       ... name of parameter to be set                          */
     /*   type_args ... contains types of arguments. Possible values:        */
     /*                   't'   for single token                             */
     /*                   'L'   for a list                                   */
     /*   args      ... array of pointers to argument strings                */
     /*   set       ... pointer to UNURAN function for set call              */
     /*                                                                      */
     /* return:                                                              */
     /*   return code from set call, i.e.                                    */
     /*      UNUR_SUCCESS ... on success                                     */
     /*      error code   ... on error                                       */
     /*----------------------------------------------------------------------*/
{
  int iarg;

  if ( !strcmp(type_args, "t") ) {
    iarg = _unur_atoi( args[0] );
  }
  else if ( strlen(type_args) == 0 ) {
    iarg = 1;
  }
  else {
    _unur_error_args(key);
#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"!");
#endif
    return UNUR_ERR_STR_INVALID;
  }
  
#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
    _unur_str_debug_set(2,key,"i",iarg);
#endif

  return set(distr,iarg);

} /* end of _unur_str_distr_set_i() */

/*---------------------------------------------------------------------------*/

int
_unur_str_distr_set_ii (UNUR_DISTR *distr, const char *key, char *type_args, char **args, distr_set_ii set)
     /*----------------------------------------------------------------------*/
     /* Set parameter in given distribution object.                          */
     /* The list of arguments and their types are given as string.           */
     /*                                                                      */
     /* type: "ii"   (int,int)                                               */
     /*                                                                      */
     /* input: "tt"  (exactly two single tokens)                             */
     /*        "L"   (exactly one list)                                      */
     /*                                                                      */
     /* action:                                                              */
     /*    if "tt": convert both to int.                                     */
     /*             execute set call.                                        */
     /*    if "L":  convert list entries to int.                             */
     /*             execute set call with first two entries.                 */
     /*                (error if list has less than two entries)             */
     /*    otherwise: error                                                  */
     /*                                                                      */
     /* parameters:                                                          */
     /*   distr     ... pointer to distribution object                       */
     /*   key       ... name of parameter to be set                          */
     /*   type_args ... contains types of arguments. Possible values:        */
     /*                   't'   for single token                             */
     /*                   'L'   for a list                                   */
     /*   args      ... array of pointers to argument strings                */
     /*   set       ... pointer to UNURAN function for set call              */
     /*                                                                      */
     /* return:                                                              */
     /*   return code from set call, i.e.                                    */
     /*      UNUR_SUCCESS ... on success                                     */
     /*      error code   ... on error                                       */
     /*----------------------------------------------------------------------*/
{
  int *iarray = NULL;
  int iarg[2];
  int result;

  /* either exactly two arguments are given */
  if ( !strcmp(type_args, "tt") ) {
    iarg[0] = _unur_atoi( args[0] );
    iarg[1] = _unur_atoi( args[1] );

#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"ii",iarg[0],iarg[1]);
#endif

    return set( distr,iarg[0],iarg[1] );
  }

  /* or one list with 2 entries is given */
  else if ( !strcmp(type_args, "L") ) {
    if ( _unur_parse_ilist( args[0], &iarray ) < 2 ) {
      _unur_error_args(key);
      free (iarray);
#ifdef UNUR_ENABLE_LOGGING
      /* write info into log file */
      if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
	_unur_str_debug_set(2,key,"!");
#endif
      return UNUR_ERR_STR_INVALID;
    }
    result = set( distr,iarray[0],iarray[1] );

#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"ii",iarray[0],iarray[1] );
#endif

    free (iarray);
    return result;
  }
  
  else {
    _unur_error_args(key);
#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"!");
#endif
    return UNUR_ERR_STR_INVALID;
  }

} /* end of _unur_str_distr_set_ii() */

/*---------------------------------------------------------------------------*/

int
_unur_str_distr_set_d (UNUR_DISTR *distr, const char *key, char *type_args, char **args, distr_set_d set)
     /*----------------------------------------------------------------------*/
     /* Set parameter in given distribution object.                          */
     /* The list of arguments and their types are given as string.           */
     /*                                                                      */
     /* type: "d"   (double)                                                 */
     /*                                                                      */
     /* input: "t"  (exactly one single token)                               */
     /*                                                                      */
     /* action:                                                              */
     /*    convert token to double.                                          */
     /*    execute set call.                                                 */
     /*    error if input != "t"                                             */
     /*                                                                      */
     /* parameters:                                                          */
     /*   distr     ... pointer to distribution object                       */
     /*   key       ... name of parameter to be set                          */
     /*   type_args ... contains types of arguments. Possible values:        */
     /*                   't'   for single token                             */
     /*                   'L'   for a list                                   */
     /*   args      ... array of pointers to argument strings                */
     /*   set       ... pointer to UNURAN function for set call              */
     /*                                                                      */
     /* return:                                                              */
     /*   return code from set call, i.e.                                    */
     /*      UNUR_SUCCESS ... on success                                     */
     /*      error code   ... on error                                       */
     /*----------------------------------------------------------------------*/
{
  double darg;

  if ( strcmp(type_args, "t") ) {
    _unur_error_args(key);
#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"!");
#endif
    return UNUR_ERR_STR_INVALID;
  }
  
  darg = _unur_atod( args[0] );

#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
    _unur_str_debug_set(2,key,"d",darg);
#endif

  return set(distr,darg);

} /* end of _unur_str_distr_set_d() */

/*---------------------------------------------------------------------------*/

int
_unur_str_distr_set_dd (UNUR_DISTR *distr, const char *key, char *type_args, char **args, distr_set_dd set)
     /*----------------------------------------------------------------------*/
     /* Set parameter in given distribution object.                          */
     /* The list of arguments and their types are given as string.           */
     /*                                                                      */
     /* type: "dd"   (double,double)                                         */
     /*                                                                      */
     /* input: "tt"  (exactly two single tokens)                             */
     /*        "L"   (exactly one list)                                      */
     /*                                                                      */
     /* action:                                                              */
     /*    if "tt": convert both to double.                                  */
     /*             execute set call.                                        */
     /*    if "L":  convert list entries to double.                          */
     /*             execute set call with first two entries.                 */
     /*                (error if list has less than two entries)             */
     /*    otherwise: error                                                  */
     /*                                                                      */
     /* parameters:                                                          */
     /*   distr     ... pointer to distribution object                       */
     /*   key       ... name of parameter to be set                          */
     /*   type_args ... contains types of arguments. Possible values:        */
     /*                   't'   for single token                             */
     /*                   'L'   for a list                                   */
     /*   args      ... array of pointers to argument strings                */
     /*   set       ... pointer to UNURAN function for set call              */
     /*                                                                      */
     /* return:                                                              */
     /*   return code from set call, i.e.                                    */
     /*      UNUR_SUCCESS ... on success                                     */
     /*      error code   ... on error                                       */
     /*----------------------------------------------------------------------*/
{
  double *darray = NULL;
  double darg[2];
  int result;

  /* either exactly two arguments are given */
  if ( !strcmp(type_args, "tt") ) {
    darg[0] = _unur_atod( args[0] );
    darg[1] = _unur_atod( args[1] );

#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"dd",darg[0],darg[1]);
#endif

    return set( distr,darg[0],darg[1] );
  }

  /* or one list with 2 entries is given */
  else if ( !strcmp(type_args, "L") ) {
    if ( _unur_parse_dlist( args[0], &darray ) < 2 ) {
      _unur_error_args(key);
      free (darray);
#ifdef UNUR_ENABLE_LOGGING
      /* write info into log file */
      if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
	_unur_str_debug_set(2,key,"!");
#endif
      return UNUR_ERR_STR_INVALID;
    }
    result = set( distr,darray[0],darray[1] );

#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"dd",darray[0],darray[1] );
#endif

    free (darray);
    return result;
  }
  
  else {
    _unur_error_args(key);
#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"!");
#endif
    return UNUR_ERR_STR_INVALID;
  }

} /* end of _unur_str_distr_set_dd() */

/*---------------------------------------------------------------------------*/

int
_unur_str_distr_set_Di (UNUR_DISTR *distr, const char *key, char *type_args, char **args, distr_set_Di set)
     /*----------------------------------------------------------------------*/
     /* Set parameter in given distribution object.                          */
     /* The list of arguments and their types are given as string.           */
     /*                                                                      */
     /* type: "Di"   (array of doubles, int)                                 */
     /*                                                                      */
     /* input: "Lt"  (list & single token)                                   */
     /*        "L"   (exactly one list)                                      */
     /*                                                                      */
     /* action:                                                              */
     /*    if "L":  convert list to array of doubles.                        */
     /*             execute set with size if array as second argument.       */
     /*    if "Lt": convert list to array of doubles.                        */
     /*             convert second token to int.                             */
     /*             set int to min of int and size of array.                 */
     /*             execute set call with int and array.                     */
     /*    otherwise: error                                                  */
     /*                                                                      */
     /* parameters:                                                          */
     /*   distr     ... pointer to distribution object                       */
     /*   key       ... name of parameter to be set                          */
     /*   type_args ... contains types of arguments. Possible values:        */
     /*                   't'   for single token                             */
     /*                   'L'   for a list                                   */
     /*   args      ... array of pointers to argument strings                */
     /*   set       ... pointer to UNURAN function for set call              */
     /*                                                                      */
     /* return:                                                              */
     /*   return code from set call, i.e.                                    */
     /*      UNUR_SUCCESS ... on success                                     */
     /*      error code   ... on error                                       */
     /*----------------------------------------------------------------------*/
{
  int result;
  int t_size;
  int size = -1;
  double *darray = NULL;

  /* either an integer and a list are given */
  if ( !strcmp(type_args, "Lt") ) {
    t_size = _unur_atoi( args[1] );
    size = _unur_parse_dlist( args[0], &darray );
    if (size > t_size)
      size = t_size;
  }
    
  /* or only a list is given */
  else if ( !strcmp(type_args, "L") ) {
    size = _unur_parse_dlist( args[0], &darray );
  }

  /* or there is a syntax error */
  if ( !(size>0) ) {
    _unur_error_args(key);
#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"!");
#endif
    result = UNUR_ERR_STR_INVALID;
  }

  else {
    /* execute set command */
    result = set( distr,darray,size );
#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"Di",darray,size,size);
#endif
  }

  if (darray) free (darray);

  return result;

} /* end of _unur_str_distr_set_Di() */

/*---------------------------------------------------------------------------*/

int
_unur_str_distr_set_C (UNUR_DISTR *distr, const char *key, char *type_args, char **args, distr_set_C set)
     /*----------------------------------------------------------------------*/
     /* Set parameter in given distribution object.                          */
     /* The list of arguments and their types are given as string.           */
     /*                                                                      */
     /* type: "C"   (string = array of char)                                 */
     /*                                                                      */
     /* input: "s"  (exactly one string)                                     */
     /*                                                                      */
     /* action:                                                              */
     /*    execute set call.                                                 */
     /*    error if input != "s"                                             */
     /*                                                                      */
     /* parameters:                                                          */
     /*   distr     ... pointer to distribution object                       */
     /*   key       ... name of parameter to be set                          */
     /*   type_args ... contains types of arguments. Possible values:        */
     /*                   's'   for single string                            */
     /*   args      ... array of pointers to argument strings                */
     /*   set       ... pointer to UNURAN function for set call              */
     /*                                                                      */
     /* return:                                                              */
     /*   return code from set call, i.e.                                    */
     /*      UNUR_SUCCESS ... on success                                     */
     /*      error code   ... on error                                       */
     /*----------------------------------------------------------------------*/
{
  char *string;

  if ( strcmp(type_args, "s") ) {
    _unur_error_args(key);
#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"!");
#endif
    return UNUR_ERR_STR_INVALID;
  }

  /* get string */
  string = args[0];
  
#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
    _unur_str_debug_set(2,key,"C",string);
#endif

  return set(distr,string);

} /* end of _unur_str_distr_set_C() */

/*---------------------------------------------------------------------------*/

/*****************************************************************************/
/**  Methods                                                                **/
/*****************************************************************************/

/*---------------------------------------------------------------------------*/

struct unur_par *
_unur_str_par( char *str_method, const UNUR_DISTR *distr, struct unur_slist *mlist )
     /*----------------------------------------------------------------------*/
     /* get parameter object for distribution and method                     */
     /*                                                                      */
     /* parameters:                                                          */
     /*   str_method ... string that contains method description             */
     /*   distr      ... pointer to distribution object                      */
     /*   mlist      ... list of allocated memory blocks                     */
     /*                                                                      */
     /* return:                                                              */
     /*   default parameters (pointer to structure)                          */
     /*                                                                      */
     /* error:                                                               */
     /*   return NULL                                                        */
     /*----------------------------------------------------------------------*/
{
  struct unur_par *par = NULL;     /* pointer to parameter object */

  char *token;             /* pointer to token in string */
  char *next;              /* pointer to next token in string */
  char *key, *value;       /* the key and its value */

  /* tokenize the string using ';' as separator and handle the tokens */
  for ( token = next = str_method;
	next != NULL && *token != '\0';
	token = next ) {

    /* find end of token (i.e. the pointer to next token) */
    next = strchr(token,';');
    if (next != NULL) {
      /* next separator found */
      *next = '\0';     /* terminate token string    */
      next++;           /* set pointer to next token */
    }

    /* token points to a key */
    key = token;

    /* get the value from the pair key=value and terminate key string */
    value = strchr(key, '=');
    if (value != NULL) {
      *value = '\0';    /* terminate key string        */
      value++;          /* set pointer to value string */
    }

    /* get (default) parameter object for method */
    if ( !strcmp( key, "method") ) {
      if (key == str_method) {
	/* "method" must be the first key.                                */
	/* (on the other hand, the first keyword always must be "method"  */
	/* since otherwise we had not called this subroutine.             */
	/* hence, we only check, whether the keyword "method" is unique.) */
	par = _unur_str_par_new(value,distr);
	if (par == NULL) {
	  /* error */
	  _unur_error(GENTYPE,UNUR_ERR_STR,"setting method failed"); 
	  return NULL;
	}
      }
      else {
	if (par) free (par);
	_unur_error(GENTYPE,UNUR_ERR_STR_SYNTAX,"method must be first key in block"); 
	return NULL;
      }
    }

    /* set parameters for method */
    else {
      if ( !_unur_str_par_set(par, key, value, mlist) ) {
	; /* error: */
      }
    }
  }

  /* end */
  return par;

} /* end of _unur_str_par() */

/*---------------------------------------------------------------------------*/

struct unur_par *
_unur_str_par_new( const char *method, const UNUR_DISTR *distr )
     /*----------------------------------------------------------------------*/
     /* get new parameter object for method                                  */
     /*                                                                      */
     /* parameters:                                                          */
     /*   method ... string that contains method name                        */
     /*   distr  ... pointer to distribution object                          */
     /*                                                                      */
     /* return:                                                              */
     /*   default parameters (pointer to structure)                          */
     /*                                                                      */
     /* error:                                                               */
     /*   return NULL                                                        */
     /*----------------------------------------------------------------------*/
{
  struct unur_par *par = NULL;
  char method_unknown;
  
#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
    _unur_str_debug_string(1,"method",method);
#endif

  /* mark method as unknown (this is a very ugly hack) */
  par = (struct unur_par *) &method_unknown;

	 switch (*method) {
	 case 'a':
		 if ( !strcmp( method, "arou") ) {
			 par = unur_arou_new(distr);
			 break;
		 }
		 if ( !strcmp( method, "auto") ) {
			 par = unur_auto_new(distr);
			 break;
		 }
		 break;
	 case 'c':
		 if ( !strcmp( method, "cstd") ) {
			 par = unur_cstd_new(distr);
			 break;
		 }
		 break;
	 case 'd':
		 if ( !strcmp( method, "dari") ) {
			 par = unur_dari_new(distr);
			 break;
		 }
		 if ( !strcmp( method, "dau") ) {
			 par = unur_dau_new(distr);
			 break;
		 }
		 if ( !strcmp( method, "dgt") ) {
			 par = unur_dgt_new(distr);
			 break;
		 }
		 if ( !strcmp( method, "dsrou") ) {
			 par = unur_dsrou_new(distr);
			 break;
		 }
		 if ( !strcmp( method, "dss") ) {
			 par = unur_dss_new(distr);
			 break;
		 }
		 if ( !strcmp( method, "dstd") ) {
			 par = unur_dstd_new(distr);
			 break;
		 }
		 break;
	 case 'e':
		 if ( !strcmp( method, "empk") ) {
			 par = unur_empk_new(distr);
			 break;
		 }
		 if ( !strcmp( method, "empl") ) {
			 par = unur_empl_new(distr);
			 break;
		 }
		 break;
	 case 'h':
		 if ( !strcmp( method, "hinv") ) {
			 par = unur_hinv_new(distr);
			 break;
		 }
		 if ( !strcmp( method, "hrb") ) {
			 par = unur_hrb_new(distr);
			 break;
		 }
		 if ( !strcmp( method, "hrd") ) {
			 par = unur_hrd_new(distr);
			 break;
		 }
		 if ( !strcmp( method, "hri") ) {
			 par = unur_hri_new(distr);
			 break;
		 }
		 break;
	 case 'm':
		 if ( !strcmp( method, "mcorr") ) {
			 par = unur_mcorr_new(distr);
			 break;
		 }
		 break;
	 case 'n':
		 if ( !strcmp( method, "ninv") ) {
			 par = unur_ninv_new(distr);
			 break;
		 }
		 if ( !strcmp( method, "nrou") ) {
			 par = unur_nrou_new(distr);
			 break;
		 }
		 break;
	 case 's':
		 if ( !strcmp( method, "srou") ) {
			 par = unur_srou_new(distr);
			 break;
		 }
		 if ( !strcmp( method, "ssr") ) {
			 par = unur_ssr_new(distr);
			 break;
		 }
		 break;
	 case 't':
		 if ( !strcmp( method, "tabl") ) {
			 par = unur_tabl_new(distr);
			 break;
		 }
		 if ( !strcmp( method, "tdr") ) {
			 par = unur_tdr_new(distr);
			 break;
		 }
		 break;
	 case 'u':
		 if ( !strcmp( method, "unif") ) {
			 par = unur_unif_new(distr);
			 break;
		 }
		 if ( !strcmp( method, "utdr") ) {
			 par = unur_utdr_new(distr);
			 break;
		 }
		 break;
	 case 'v':
		 if ( !strcmp( method, "vempk") ) {
			 par = unur_vempk_new(distr);
			 break;
		 }
		 if ( !strcmp( method, "vmt") ) {
			 par = unur_vmt_new(distr);
			 break;
		 }
		 if ( !strcmp( method, "vnrou") ) {
			 par = unur_vnrou_new(distr);
			 break;
		 }
	 }

   if (par == (struct unur_par *) &method_unknown) {
     /* unknown method */
     _unur_error_unknown(method,"method");
     par = NULL;
   }
   else if (par == NULL) {
     /* invalid data for chosen method */
     _unur_error_invalid(method,"method");
   }

   return par;
} /* end of _unur_str_par_new() */

/*---------------------------------------------------------------------------*/

int
_unur_str_par_set( UNUR_PAR *par, const char *key, char *value, struct unur_slist *mlist )
     /*----------------------------------------------------------------------*/
     /* set parameters for method                                            */
     /*                                                                      */
     /* parameters:                                                          */
     /*   par   ... pointer to parameter for building generator object       */
     /*   key   ... string that contains key for parameter                   */
     /*   value ... string that contains list of arguments for parameter     */
     /*   mlist ... list of allocated memory blocks                          */
     /*                                                                      */
     /* return:                                                              */
     /*   UNUR_SUCCESS ... on success                                        */
     /*   error code   ... on error                                          */
     /*----------------------------------------------------------------------*/
{
  int result = 0;                  /* result of UNURAN set call (0 or 1) */

  /* storing arguments of set calls: */
  char type_args[MAX_SET_ARGS+1];  /* array containing info about arguments */
  char *args[MAX_SET_ARGS+1];      /* pointers to argument strings */

  /* tokenize argument string */
  if (_unur_str_set_args( value, type_args, args, MAX_SET_ARGS ) < 0) {
    /* error */
    return UNUR_ERR_STR_SYNTAX;
  }

  /* set result indicator to unknown */
  result = UNUR_ERR_STR_UNKNOWN;

  /* find and execute set call */
	 switch (par->method) {
	 case UNUR_METH_AROU:
		 switch (*key) {
		 case 'c':
			 if ( !strcmp(key, "center") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double center */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_arou_set_center);
				 break;
			 }
			 if ( !strcmp(key, "cpoints") ) {
				 /* n = 2; type = iD:  UNUR_PAR *parameters, int n_stp, const double *stp */
				 result = _unur_str_par_set_iD(par,key,type_args,args,unur_arou_set_cpoints,mlist);
				 break;
			 }
			 break;
		 case 'd':
			 if ( !strcmp(key, "darsfactor") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double factor */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_arou_set_darsfactor);
				 break;
			 }
			 break;
		 case 'g':
			 if ( !strcmp(key, "guidefactor") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double factor */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_arou_set_guidefactor);
				 break;
			 }
			 break;
		 case 'm':
			 if ( !strcmp(key, "max_segments") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int max_segs */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_arou_set_max_segments);
				 break;
			 }
			 if ( !strcmp(key, "max_sqhratio") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double max_ratio */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_arou_set_max_sqhratio);
				 break;
			 }
			 break;
		 case 'p':
			 if ( !strcmp(key, "pedantic") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int pedantic */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_arou_set_pedantic);
				 break;
			 }
			 break;
		 case 'u':
			 if ( !strcmp(key, "usecenter") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int usecenter */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_arou_set_usecenter);
				 break;
			 }
			 if ( !strcmp(key, "usedars") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int usedars */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_arou_set_usedars);
				 break;
			 }
			 break;
		 case 'v':
			 if ( !strcmp(key, "verify") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int verify */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_arou_set_verify);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_AUTO:
		 switch (*key) {
		 case 'l':
			 if ( !strcmp(key, "logss") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int logss */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_auto_set_logss);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_CSTD:
		 switch (*key) {
		 case 'v':
			 if ( !strcmp(key, "variant") ) {
				 /* n = 1; type = u:  UNUR_PAR *parameters, unsigned variant */
				 result = _unur_str_par_set_u(par,key,type_args,args,unur_cstd_set_variant);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_DARI:
		 switch (*key) {
		 case 'c':
			 if ( !strcmp(key, "cpfactor") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double cp_factor */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_dari_set_cpfactor);
				 break;
			 }
			 break;
		 case 's':
			 if ( !strcmp(key, "squeeze") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int squeeze */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_dari_set_squeeze);
				 break;
			 }
			 break;
		 case 't':
			 if ( !strcmp(key, "tablesize") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int size */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_dari_set_tablesize);
				 break;
			 }
			 break;
		 case 'v':
			 if ( !strcmp(key, "verify") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int verify */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_dari_set_verify);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_DAU:
		 switch (*key) {
		 case 'u':
			 if ( !strcmp(key, "urnfactor") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double factor */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_dau_set_urnfactor);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_DGT:
		 switch (*key) {
		 case 'g':
			 if ( !strcmp(key, "guidefactor") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double factor */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_dgt_set_guidefactor);
				 break;
			 }
			 break;
		 case 'v':
			 if ( !strcmp(key, "variant") ) {
				 /* n = 1; type = u:  UNUR_PAR *parameters, unsigned variant */
				 result = _unur_str_par_set_u(par,key,type_args,args,unur_dgt_set_variant);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_DSROU:
		 switch (*key) {
		 case 'c':
			 if ( !strcmp(key, "cdfatmode") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double Fmode */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_dsrou_set_cdfatmode);
				 break;
			 }
			 break;
		 case 'v':
			 if ( !strcmp(key, "verify") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int verify */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_dsrou_set_verify);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_DSTD:
		 switch (*key) {
		 case 'v':
			 if ( !strcmp(key, "variant") ) {
				 /* n = 1; type = u:  UNUR_PAR *parameters, unsigned variant */
				 result = _unur_str_par_set_u(par,key,type_args,args,unur_dstd_set_variant);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_EMPK:
		 switch (*key) {
		 case 'b':
			 if ( !strcmp(key, "beta") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double beta */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_empk_set_beta);
				 break;
			 }
			 break;
		 case 'k':
			 if ( !strcmp(key, "kernel") ) {
				 /* n = 1; type = u:  UNUR_PAR *parameters, unsigned kernel*/
				 result = _unur_str_par_set_u(par,key,type_args,args,unur_empk_set_kernel);
				 break;
			 }
			 break;
		 case 'p':
			 if ( !strcmp(key, "positive") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int positive */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_empk_set_positive);
				 break;
			 }
			 break;
		 case 's':
			 if ( !strcmp(key, "smoothing") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double smoothing */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_empk_set_smoothing);
				 break;
			 }
			 break;
		 case 'v':
			 if ( !strcmp(key, "varcor") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int varcor */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_empk_set_varcor);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_HINV:
		 switch (*key) {
		 case 'b':
			 if ( !strcmp(key, "boundary") ) {
				 /* n = 2; type = dd:  UNUR_PAR *parameters, double left, double right */
				 result = _unur_str_par_set_dd(par,key,type_args,args,unur_hinv_set_boundary);
				 break;
			 }
			 break;
		 case 'c':
			 if ( !strcmp(key, "cpoints") ) {
				 /* n = 2; type = Di:  UNUR_PAR *parameters, const double *stp, int n_stp */
				 result = _unur_str_par_set_Di(par,key,type_args,args,unur_hinv_set_cpoints,mlist);
				 break;
			 }
			 break;
		 case 'g':
			 if ( !strcmp(key, "guidefactor") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double factor */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_hinv_set_guidefactor);
				 break;
			 }
			 break;
		 case 'm':
			 if ( !strcmp(key, "max_intervals") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int max_ivs */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_hinv_set_max_intervals);
				 break;
			 }
			 break;
		 case 'o':
			 if ( !strcmp(key, "order") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int order*/
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_hinv_set_order);
				 break;
			 }
			 break;
		 case 'u':
			 if ( !strcmp(key, "u_resolution") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double u_resolution*/
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_hinv_set_u_resolution);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_HRB:
		 switch (*key) {
		 case 'u':
			 if ( !strcmp(key, "upperbound") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double upperbound */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_hrb_set_upperbound);
				 break;
			 }
			 break;
		 case 'v':
			 if ( !strcmp(key, "verify") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int verify */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_hrb_set_verify);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_HRD:
		 switch (*key) {
		 case 'v':
			 if ( !strcmp(key, "verify") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int verify */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_hrd_set_verify);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_HRI:
		 switch (*key) {
		 case 'p':
			 if ( !strcmp(key, "p0") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double p0 */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_hri_set_p0);
				 break;
			 }
			 break;
		 case 'v':
			 if ( !strcmp(key, "verify") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int verify */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_hri_set_verify);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_NINV:
		 switch (*key) {
		 case 'm':
			 if ( !strcmp(key, "max_iter") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int max_iter */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_ninv_set_max_iter);
				 break;
			 }
			 break;
		 case 's':
			 if ( !strcmp(key, "start") ) {
				 /* n = 2; type = dd:  UNUR_PAR *parameters, double left, double right*/
				 result = _unur_str_par_set_dd(par,key,type_args,args,unur_ninv_set_start);
				 break;
			 }
			 break;
		 case 't':
			 if ( !strcmp(key, "table") ) {
				 /* n = 1; type = i: UNUR_PAR *parameters, int no_of_points*/
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_ninv_set_table);
				 break;
			 }
			 break;
		 case 'u':
			 if ( !strcmp(key, "usenewton") ) {
				 /* n = 0; type = :  UNUR_PAR *parameters */
				 result = _unur_str_par_set_void(par,key,type_args,args,unur_ninv_set_usenewton);
				 break;
			 }
			 if ( !strcmp(key, "useregula") ) {
				 /* n = 0; type = :  UNUR_PAR *parameters */
				 result = _unur_str_par_set_void(par,key,type_args,args,unur_ninv_set_useregula);
				 break;
			 }
			 break;
		 case 'x':
			 if ( !strcmp(key, "x_resolution") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double x_resolution*/
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_ninv_set_x_resolution);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_NROU:
		 switch (*key) {
		 case 'c':
			 if ( !strcmp(key, "center") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double center */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_nrou_set_center);
				 break;
			 }
			 break;
		 case 'u':
			 if ( !strcmp(key, "u") ) {
				 /* n = 2; type = dd:  UNUR_PAR *parameters, double umin, double umax */
				 result = _unur_str_par_set_dd(par,key,type_args,args,unur_nrou_set_u);
				 break;
			 }
			 break;
		 case 'v':
			 if ( !strcmp(key, "v") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double vmax */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_nrou_set_v);
				 break;
			 }
			 if ( !strcmp(key, "verify") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int verify */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_nrou_set_verify);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_SROU:
		 switch (*key) {
		 case 'c':
			 if ( !strcmp(key, "cdfatmode") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double Fmode */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_srou_set_cdfatmode);
				 break;
			 }
			 break;
		 case 'p':
			 if ( !strcmp(key, "pdfatmode") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double fmode */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_srou_set_pdfatmode);
				 break;
			 }
			 break;
		 case 'r':
			 if ( !strcmp(key, "r") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double r */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_srou_set_r);
				 break;
			 }
			 break;
		 case 'u':
			 if ( !strcmp(key, "usemirror") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int usemirror */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_srou_set_usemirror);
				 break;
			 }
			 if ( !strcmp(key, "usesqueeze") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int usesqueeze */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_srou_set_usesqueeze);
				 break;
			 }
			 break;
		 case 'v':
			 if ( !strcmp(key, "verify") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int verify */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_srou_set_verify);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_SSR:
		 switch (*key) {
		 case 'c':
			 if ( !strcmp(key, "cdfatmode") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double Fmode */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_ssr_set_cdfatmode);
				 break;
			 }
			 break;
		 case 'p':
			 if ( !strcmp(key, "pdfatmode") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double fmode */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_ssr_set_pdfatmode);
				 break;
			 }
			 break;
		 case 'u':
			 if ( !strcmp(key, "usesqueeze") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int usesqueeze */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_ssr_set_usesqueeze);
				 break;
			 }
			 break;
		 case 'v':
			 if ( !strcmp(key, "verify") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int verify */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_ssr_set_verify);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_TABL:
		 switch (*key) {
		 case 'a':
			 if ( !strcmp(key, "areafraction") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double fraction */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_tabl_set_areafraction);
				 break;
			 }
			 break;
		 case 'b':
			 if ( !strcmp(key, "boundary") ) {
				 /* n = 2; type = dd:  UNUR_PAR *parameters, double left, double right */
				 result = _unur_str_par_set_dd(par,key,type_args,args,unur_tabl_set_boundary);
				 break;
			 }
			 break;
		 case 'd':
			 if ( !strcmp(key, "darsfactor") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double factor */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_tabl_set_darsfactor);
				 break;
			 }
			 break;
		 case 'g':
			 if ( !strcmp(key, "guidefactor") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double factor */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_tabl_set_guidefactor);
				 break;
			 }
			 break;
		 case 'm':
			 if ( !strcmp(key, "max_intervals") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int max_ivs */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_tabl_set_max_intervals);
				 break;
			 }
			 if ( !strcmp(key, "max_sqhratio") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double max_ratio */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_tabl_set_max_sqhratio);
				 break;
			 }
			 break;
		 case 'n':
			 if ( !strcmp(key, "nstp") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int n_stp */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_tabl_set_nstp);
				 break;
			 }
			 break;
		 case 's':
			 if ( !strcmp(key, "slopes") ) {
				 /* n = 2; type = Di:  UNUR_PAR *parameters, const double *slopes, int n_slopes */
				 result = _unur_str_par_set_Di(par,key,type_args,args,unur_tabl_set_slopes,mlist);
				 break;
			 }
			 break;
		 case 'u':
			 if ( !strcmp(key, "usedars") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int usedars */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_tabl_set_usedars);
				 break;
			 }
			 break;
		 case 'v':
			 if ( !strcmp(key, "variant_splitmode") ) {
				 /* n = 1; type = u:  UNUR_PAR *parameters, unsigned splitmode */
				 result = _unur_str_par_set_u(par,key,type_args,args,unur_tabl_set_variant_splitmode);
				 break;
			 }
			 if ( !strcmp(key, "verify") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int verify */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_tabl_set_verify);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_TDR:
		 switch (*key) {
		 case 'c':
			 if ( !strcmp(key, "c") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double c */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_tdr_set_c);
				 break;
			 }
			 if ( !strcmp(key, "center") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double center */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_tdr_set_center);
				 break;
			 }
			 if ( !strcmp(key, "cpoints") ) {
				 /* n = 2; type = iD:  UNUR_PAR *parameters, int n_stp, const double *stp */
				 result = _unur_str_par_set_iD(par,key,type_args,args,unur_tdr_set_cpoints,mlist);
				 break;
			 }
			 break;
		 case 'd':
			 if ( !strcmp(key, "darsfactor") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double factor */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_tdr_set_darsfactor);
				 break;
			 }
			 break;
		 case 'g':
			 if ( !strcmp(key, "guidefactor") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double factor */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_tdr_set_guidefactor);
				 break;
			 }
			 break;
		 case 'm':
			 if ( !strcmp(key, "max_intervals") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int max_ivs */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_tdr_set_max_intervals);
				 break;
			 }
			 if ( !strcmp(key, "max_sqhratio") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double max_ratio */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_tdr_set_max_sqhratio);
				 break;
			 }
			 break;
		 case 'p':
			 if ( !strcmp(key, "pedantic") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int pedantic */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_tdr_set_pedantic);
				 break;
			 }
			 break;
		 case 'u':
			 if ( !strcmp(key, "usecenter") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int usecenter */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_tdr_set_usecenter);
				 break;
			 }
			 if ( !strcmp(key, "usedars") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int usedars */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_tdr_set_usedars);
				 break;
			 }
			 if ( !strcmp(key, "usemode") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int usemode */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_tdr_set_usemode);
				 break;
			 }
			 break;
		 case 'v':
			 if ( !strcmp(key, "variant_gw") ) {
				 /* n = 0; type = :  UNUR_PAR *parameters */
				 result = _unur_str_par_set_void(par,key,type_args,args,unur_tdr_set_variant_gw);
				 break;
			 }
			 if ( !strcmp(key, "variant_ia") ) {
				 /* n = 0; type = :  UNUR_PAR *parameters */
				 result = _unur_str_par_set_void(par,key,type_args,args,unur_tdr_set_variant_ia);
				 break;
			 }
			 if ( !strcmp(key, "variant_ps") ) {
				 /* n = 0; type = :  UNUR_PAR *parameters */
				 result = _unur_str_par_set_void(par,key,type_args,args,unur_tdr_set_variant_ps);
				 break;
			 }
			 if ( !strcmp(key, "verify") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int verify */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_tdr_set_verify);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_UTDR:
		 switch (*key) {
		 case 'c':
			 if ( !strcmp(key, "cpfactor") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double cp_factor */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_utdr_set_cpfactor);
				 break;
			 }
			 break;
		 case 'd':
			 if ( !strcmp(key, "deltafactor") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double delta */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_utdr_set_deltafactor);
				 break;
			 }
			 break;
		 case 'p':
			 if ( !strcmp(key, "pdfatmode") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double fmode */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_utdr_set_pdfatmode);
				 break;
			 }
			 break;
		 case 'v':
			 if ( !strcmp(key, "verify") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int verify */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_utdr_set_verify);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_VEMPK:
		 switch (*key) {
		 case 's':
			 if ( !strcmp(key, "smoothing") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double smoothing */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_vempk_set_smoothing);
				 break;
			 }
			 break;
		 case 'v':
			 if ( !strcmp(key, "varcor") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int varcor */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_vempk_set_varcor);
				 break;
			 }
		 }
		 break;
	 case UNUR_METH_VNROU:
		 switch (*key) {
		 case 'r':
			 if ( !strcmp(key, "r") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double r */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_vnrou_set_r);
				 break;
			 }
			 break;
		 case 'v':
			 if ( !strcmp(key, "v") ) {
				 /* n = 1; type = d:  UNUR_PAR *parameters, double vmax */
				 result = _unur_str_par_set_d(par,key,type_args,args,unur_vnrou_set_v);
				 break;
			 }
			 if ( !strcmp(key, "verify") ) {
				 /* n = 1; type = i:  UNUR_PAR *parameters, int verify */
				 result = _unur_str_par_set_i(par,key,type_args,args,unur_vnrou_set_verify);
				 break;
			 }
		 }
		 break;
	 }

	 /* Unsupported set commands: */
 	 /* int unur_empk_set_kernelgen( UNUR_PAR *parameters, const UNUR_GEN *kernelgen, double alpha, double kernelvar );
		 n = 3; type = ?dd	 */
	 /* int unur_vnrou_set_u( UNUR_PAR *parameters, double *umin, double *umax );
		 n = 2; type = DD	 */


   if (result == UNUR_ERR_STR_UNKNOWN) {
     /* no valid parameter found */
     /* try extra keywords */
     if ( !strcmp(key, "debug") ) {
       /* n = 1; type = u:  UNUR_PAR *parameters, unsigned DEBUG */
       result = _unur_str_par_set_u(par,key,type_args,args,unur_set_debug);
     }
   }
  
  /* check result */
  if (result == UNUR_ERR_STR_UNKNOWN) {
    /* unknown parameter */
    _unur_error_unknown(key,"parameter for given method");
    return UNUR_ERR_STR_UNKNOWN;
  }
  else if (result != UNUR_SUCCESS) {
    /* invalid data for parameter */
    _unur_error_invalid(key,"set call");
    return UNUR_ERR_STR_SYNTAX;
  }
  
  return UNUR_SUCCESS;

} /* end of _unur_str_par_set() */

/*---------------------------------------------------------------------------*/

int
_unur_str_par_set_void (UNUR_PAR *par, const char *key, char *type_args, char **args, par_set_void set)
     /*----------------------------------------------------------------------*/
     /* Set parameter in given parameter object.                             */
     /* The list of arguments and their types are given as string.           */
     /*                                                                      */
     /* type: void   (no arguments required)                                 */
     /*                                                                      */
     /* input: void  (no arguments should be given)                          */
     /*                                                                      */
     /* action:                                                              */
     /*    execute set call.                                                 */
     /*    warning if any argument is given.                                 */
     /*                                                                      */
     /* parameters:                                                          */
     /*   par       ... pointer to parameter object                          */
     /*   key       ... name of parameter to be set                          */
     /*   type_args ... contains types of arguments. Possible values:        */
     /*                   't'   for single token                             */
     /*                   'L'   for a list                                   */
     /*   args      ... array of pointers to argument strings                */
     /*   set       ... pointer to UNURAN function for set call              */
     /*                                                                      */
     /* return:                                                              */
     /*   return code from set call, i.e.                                    */
     /*      UNUR_SUCCESS ... on success                                     */
     /*      error code   ... on error                                       */
     /*----------------------------------------------------------------------*/
{

#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
    _unur_str_debug_set(2,key,"v");
#endif

  if (*type_args != '\0')
    _unur_error_args(key);

  return set(par);

} /* end of _unur_str_par_set_void() */

/*---------------------------------------------------------------------------*/

int
_unur_str_par_set_i (UNUR_PAR *par, const char *key, char *type_args, char **args, par_set_i set)
     /*----------------------------------------------------------------------*/
     /* Set parameter in given parameter object.                             */
     /* The list of arguments and their types are given as string.           */
     /*                                                                      */
     /* type: "i"   (int)                                                    */
     /*                                                                      */
     /* input: "t"   (exactly one single token)                              */
     /*         void (no argument given)                                     */
     /*                                                                      */
     /* action:                                                              */
     /*    convert token to int.                                             */
     /*      (if no argument is given, use 1.)                               */
     /*    execute set call.                                                 */
     /*    error if input != "t"                                             */
     /*                                                                      */
     /* parameters:                                                          */
     /*   par       ... pointer to parameter object                          */
     /*   key       ... name of parameter to be set                          */
     /*   type_args ... contains types of arguments. Possible values:        */
     /*                   't'   for single token                             */
     /*                   'L'   for a list                                   */
     /*   args      ... array of pointers to argument strings                */
     /*   set       ... pointer to UNURAN function for set call              */
     /*                                                                      */
     /* return:                                                              */
     /*   return code from set call, i.e.                                    */
     /*      UNUR_SUCCESS ... on success                                     */
     /*      error code   ... on error                                       */
     /*----------------------------------------------------------------------*/
{
  int iarg;

  if ( !strcmp(type_args, "t") ) {
    iarg = _unur_atoi( args[0] );
  }
  else if ( strlen(type_args) == 0 ) {
    iarg = 1;
  }
  else {
    _unur_error_args(key);
#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"!");
#endif
    return UNUR_ERR_STR_INVALID;
  }
  
#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
    _unur_str_debug_set(2,key,"i",iarg);
#endif

  return set(par,iarg);

} /* end of _unur_str_par_set_i() */

/*---------------------------------------------------------------------------*/

int
_unur_str_par_set_u (UNUR_PAR *par, const char *key, char *type_args, char **args, par_set_u set)
     /*----------------------------------------------------------------------*/
     /* Set parameter in given parameter object.                             */
     /* The list of arguments and their types are given as string.           */
     /*                                                                      */
     /* type: "u"   (unsigned)                                               */
     /*                                                                      */
     /* input: "t"  (exactly one single token)                               */
     /*                                                                      */
     /* action:                                                              */
     /*    convert token to unsigned.                                        */
     /*    execute set call.                                                 */
     /*    error if input != "t"                                             */
     /*                                                                      */
     /* parameters:                                                          */
     /*   par       ... pointer to parameter object                          */
     /*   key       ... name of parameter to be set                          */
     /*   type_args ... contains types of arguments. Possible values:        */
     /*                   't'   for single token                             */
     /*                   'L'   for a list                                   */
     /*   args      ... array of pointers to argument strings                */
     /*   set       ... pointer to UNURAN function for set call              */
     /*                                                                      */
     /* return:                                                              */
     /*   return code from set call, i.e.                                    */
     /*      UNUR_SUCCESS ... on success                                     */
     /*      error code   ... on error                                       */
     /*----------------------------------------------------------------------*/
{
  unsigned uarg;

  if ( strcmp(type_args, "t") ) {
    _unur_error_args(key);
#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"!");
#endif
    return UNUR_ERR_STR_INVALID;
  }
  
  uarg = _unur_atou( args[0] );

#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
    _unur_str_debug_set(2,key,"u",uarg);
#endif

  return set(par,uarg);

} /* end of _unur_str_par_set_u() */

/*---------------------------------------------------------------------------*/

int
_unur_str_par_set_d (UNUR_PAR *par, const char *key, char *type_args, char **args, par_set_d set)
     /*----------------------------------------------------------------------*/
     /* Set parameter in given parameter object.                             */
     /* The list of arguments and their types are given as string.           */
     /*                                                                      */
     /* type: "d"   (double)                                                 */
     /*                                                                      */
     /* input: "t"  (exactly one single token)                               */
     /*                                                                      */
     /* action:                                                              */
     /*    convert token to double.                                          */
     /*    execute set call.                                                 */
     /*    error if input != "t"                                             */
     /*                                                                      */
     /* parameters:                                                          */
     /*   par       ... pointer to parameter object                          */
     /*   key       ... name of parameter to be set                          */
     /*   type_args ... contains types of arguments. Possible values:        */
     /*                   't'   for single token                             */
     /*                   'L'   for a list                                   */
     /*   args      ... array of pointers to argument strings                */
     /*   set       ... pointer to UNURAN function for set call              */
     /*                                                                      */
     /* return:                                                              */
     /*   return code from set call, i.e.                                    */
     /*      UNUR_SUCCESS ... on success                                     */
     /*      error code   ... on error                                       */
     /*----------------------------------------------------------------------*/
{
  double darg;

  if ( strcmp(type_args, "t") ) {
    _unur_error_args(key);
#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"!");
#endif
    return UNUR_ERR_STR_INVALID;
  }
  
  darg = _unur_atod( args[0] );

#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
    _unur_str_debug_set(2,key,"d",darg);
#endif

  return set(par,darg);

} /* end of _unur_str_par_set_d() */

/*---------------------------------------------------------------------------*/

int
_unur_str_par_set_dd (UNUR_PAR *par, const char *key, char *type_args, char **args, par_set_dd set)
     /*----------------------------------------------------------------------*/
     /* Set parameter in given parameter object.                             */
     /* The list of arguments and their types are given as string.           */
     /*                                                                      */
     /* type: "dd"   (double,double)                                         */
     /*                                                                      */
     /* input: "tt"  (exactly two single tokens)                             */
     /*        "L"   (exactly one list)                                      */
     /*                                                                      */
     /* action:                                                              */
     /*    if "tt": convert both to double.                                  */
     /*             execute set call.                                        */
     /*    if "L":  convert list entries to double.                          */
     /*             execute set call with first two entries.                 */
     /*                (error if list has less than two entries)             */
     /*    otherwise: error                                                  */
     /*                                                                      */
     /* parameters:                                                          */
     /*   par       ... pointer to parameter object                          */
     /*   key       ... name of parameter to be set                          */
     /*   type_args ... contains types of arguments. Possible values:        */
     /*                   't'   for single token                             */
     /*                   'L'   for a list                                   */
     /*   args      ... array of pointers to argument strings                */
     /*   set       ... pointer to UNURAN function for set call              */
     /*                                                                      */
     /* return:                                                              */
     /*   return code from set call, i.e.                                    */
     /*      UNUR_SUCCESS ... on success                                     */
     /*      error code   ... on error                                       */
     /*----------------------------------------------------------------------*/
{
  double *darray = NULL;
  double darg[2];
  int result;

  /* either exactly two arguments are given */
  if ( !strcmp(type_args, "tt") ) {
    darg[0] = _unur_atod( args[0] );
    darg[1] = _unur_atod( args[1] );

#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"dd",darg[0],darg[1]);
#endif

    return set( par,darg[0],darg[1] );
  }

  /* or one list with 2 entries is given */
  else if ( !strcmp(type_args, "L") ) {
    if ( _unur_parse_dlist( args[0], &darray ) < 2 ) {
      _unur_error_args(key);
      free (darray);
#ifdef UNUR_ENABLE_LOGGING
      /* write info into log file */
      if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
	_unur_str_debug_set(2,key,"!");
#endif
      return UNUR_ERR_STR_INVALID;
    }
    result = set( par,darray[0],darray[1] );

#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"dd",darray[0],darray[1] );
#endif

    free (darray);
    return result;
  }
  
  else {
    _unur_error_args(key);
#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"!");
#endif
    return UNUR_ERR_STR_INVALID;
  }

} /* end of _unur_str_par_set_dd() */

/*---------------------------------------------------------------------------*/

int
_unur_str_par_set_iD (UNUR_PAR *par, const char *key, char *type_args, char **args,
		      par_set_iD set, struct unur_slist *mlist )
     /*----------------------------------------------------------------------*/
     /* Set parameter in given parameter object.                             */
     /* The list of arguments and their types are given as string.           */
     /*                                                                      */
     /* type: "iD"   (int, array of doubles)                                 */
     /*                                                                      */
     /* input: "tL"  (single token & list)                                   */
     /*        "L"   (exactly one list)                                      */
     /*        "t"   (single token)                                          */
     /*                                                                      */
     /* action:                                                              */
     /*    if "L":  convert list to array of doubles.                        */
     /*             execute set with size if array as first argument.        */
     /*    if "t":  convert token to int.                                    */ 
     /*             execute set call with NULL as its second argument.       */
     /*    if "tL": convert list to array of doubles.                        */
     /*             convert first token to int.                              */
     /*             set int to min of int and size of array.                 */
     /*             execute set call with int and array.                     */
     /*    otherwise: error                                                  */
     /*                                                                      */
     /* parameters:                                                          */
     /*   par       ... pointer to parameter object                          */
     /*   key       ... name of parameter to be set                          */
     /*   type_args ... contains types of arguments. Possible values:        */
     /*                   't'   for single token                             */
     /*                   'L'   for a list                                   */
     /*   args      ... array of pointers to argument strings                */
     /*   set       ... pointer to UNURAN function for set call              */
     /*   mlist     ... list of allocated memory blocks                      */
     /*                                                                      */
     /* return:                                                              */
     /*   return code from set call, i.e.                                    */
     /*      UNUR_SUCCESS ... on success                                     */
     /*      error code   ... on error                                       */
     /*----------------------------------------------------------------------*/
{
  int result;
  int t_size;
  int size = -1;
  double *darray = NULL;

  /* either an integer and a list are given */
  if ( !strcmp(type_args, "tL") ) {
    t_size = _unur_atoi( args[0] );
    size = _unur_parse_dlist( args[1], &darray );
    if (size > 0) {
      if (size > t_size)  size = t_size;
    }
    else {
      /* empty list */
      size = t_size;
      if (darray) free (darray);
      darray = NULL;
    }
  }

  /* or only an integer is given */
  else if ( !strcmp(type_args, "t") ) {
    size = _unur_atoi( args[0] );
    darray = NULL;
  }
  
  /* or only a list is given */
  else if ( !strcmp(type_args, "L") ) {
    size = _unur_parse_dlist( args[0], &darray );
  }
  
  /* or there is a syntax error */
  if ( !(size>0) ) {
    _unur_error_args(key);
#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"!");
#endif
    result = UNUR_ERR_STR_INVALID;
  }
  
  else {
    /* execute set command */
    result = set( par,size,darray );
#ifdef UNUR_ENABLE_LOGGING
    /* write info into log file */
    if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
      _unur_str_debug_set(2,key,"iD",size,darray,size);
#endif
  }

  /* store pointer to allocated array */
  if (darray) 
    _unur_slist_append( mlist, darray );

  return result;

} /* end of _unur_str_par_set_iD() */

/*---------------------------------------------------------------------------*/

int
_unur_str_par_set_Di (UNUR_PAR *par, const char *key, char *type_args, char **args,
		      par_set_Di set, struct unur_slist *mlist )
     /*----------------------------------------------------------------------*/
     /* Set parameter in given parameter object.                             */
     /* The list of arguments and their types are given as string.           */
     /*                                                                      */
     /* type: "Di"   (array of doubles, int)                                 */
     /*                                                                      */
     /* input: "Lt"  (list & single token & list)                            */
     /*                                                                      */
     /* action:                                                              */
     /*    convert list to array of doubles.                                 */
     /*    convert token to int.                                             */
     /*    execute set call with array and int.                              */
     /*                                                                      */
     /* parameters:                                                          */
     /*   par       ... pointer to parameter object                          */
     /*   key       ... name of parameter to be set                          */
     /*   type_args ... contains types of arguments. Possible values:        */
     /*                   't'   for single token                             */
     /*                   'L'   for a list                                   */
     /*   args      ... array of pointers to argument strings                */
     /*   set       ... pointer to UNURAN function for set call              */
     /*   mlist     ... list of allocated memory blocks                      */
     /*                                                                      */
     /* return:                                                              */
     /*   return code from set call, i.e.                                    */
     /*      UNUR_SUCCESS ... on success                                     */
     /*      error code   ... on error                                       */
     /*----------------------------------------------------------------------*/
{
  int result;
  int t_size;
  int size;
  double *darray = NULL;

  if ( !strcmp(type_args, "Lt") ) {
    t_size = _unur_atoi( args[1] );
    size = _unur_parse_dlist( args[0], &darray );
    if (size > 0) {
      result = set( par,darray,t_size );
#ifdef UNUR_ENABLE_LOGGING
      /* write info into log file */
      if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
	_unur_str_debug_set(2,key,"Di",darray,size,size);
#endif

      /* store pointer to allocated array */
      if (darray) 
	_unur_slist_append( mlist, darray );

      return result;
    }
  }

  /* there is a syntax error */
  _unur_error_args(key);
#ifdef UNUR_ENABLE_LOGGING
  /* write info into log file */
  if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
    _unur_str_debug_set(2,key,"!");
#endif

  return UNUR_ERR_STR_INVALID;
} /* end of _unur_str_par_set_Di() */

/*---------------------------------------------------------------------------*/

/*****************************************************************************/
/**  Uniform RNG                                                            **/
/*****************************************************************************/

UNUR_URNG *
_unur_str2urng( char *str_urng )
     /*----------------------------------------------------------------------*/
     /* get uniform RNG                                                      */
     /*                                                                      */
     /* parameters:                                                          */
     /*   str_urng ... string that contains description of uniform RNG       */
     /*                                                                      */
     /* return:                                                              */
     /*   pointer to uniform RNG                                             */
     /*                                                                      */
     /* The string must not contain any other keywords than "urng".          */
     /* the argument is passed unchanged to the PRNG package.                */
     /*                                                                      */
     /* error:                                                               */
     /*   return NULL                                                        */
     /*----------------------------------------------------------------------*/
{
/*---------------------------------------------------------------------------*/
#if UNUR_URNG_TYPE == UNUR_URNG_PRNG
/*---------------------------------------------------------------------------*/

  UNUR_URNG *urng = NULL;

  char *token;             /* pointer to token in string */
  char *next;              /* pointer to next token in string */
  char *key, *value;       /* the key and its value */

  /* tokenize the string using ';' as separator and handle the tokens */
  for ( token = next = str_urng;
	next != NULL && *token != '\0';
	token = next ) {

    /* find end of token (i.e. the pointer to next token) */
    next = strchr(token,';');
    if (next != NULL) {
      /* next separator found */
      *next = '\0';     /* terminate token string    */
      next++;           /* set pointer to next token */
    }
    
    /* token points to a key */
    key = token;

    /* get the value from the pair key=value and terminate key string */
    value = strchr(key, '=');
    if (value != NULL) {
      *value = '\0';    /* terminate key string        */
      value++;          /* set pointer to value string */
    }
    
    /* get (default) parameter object for method */
    if ( !strcmp( key, "urng") ) {
      if (key == str_urng) {
	/* "urng" must be the first key.                                  */
	/* (on the other hand, the first keyword always must be "urng"    */
	/* since otherwise we had not called this subroutine.             */
	/* hence, we only check, whether the keyword "urng" is unique.)   */
	urng = prng_new(value);
	if (urng == NULL) {
	  /* error */
	  _unur_error(GENTYPE,UNUR_ERR_STR,"setting URNG failed"); 
	  return NULL;
	}
	else {
#ifdef UNUR_ENABLE_LOGGING
	  /* write info into log file */
	  if (_unur_default_debugflag & UNUR_DEBUG_SETUP)
	    _unur_str_debug_string(1,"urng",value);
#endif
	  ;
	}
      }
      else {
	if (urng) prng_free (urng);
	_unur_error(GENTYPE,UNUR_ERR_STR_SYNTAX,"urng must be first key"); 
	return NULL;
      }
    }

    /* there are no other parameters */
    else {
      /* error: unknown parameter */
      if (urng) prng_free (urng);
      _unur_error_unknown(key,"parameter for uniform RNG");
      return NULL;
    }
  }

  /*return pointer to uniform RNG */
  return urng;

/*---------------------------------------------------------------------------*/
#else
/*---------------------------------------------------------------------------*/

  _unur_error(GENTYPE,UNUR_ERR_STR,"setting URNG requires PRNG library enabled"); 
  return NULL;

/*---------------------------------------------------------------------------*/
#endif
/*---------------------------------------------------------------------------*/
} /* end of _unur_str2urng() */

/*---------------------------------------------------------------------------*/

/*****************************************************************************/
/**  Misc                                                                   **/
/*****************************************************************************/

/*---------------------------------------------------------------------------*/

int
_unur_str_set_args( char *value, char *type_args, char **args, int max_args )
     /*----------------------------------------------------------------------*/
     /* parse argument string for set call                                   */
     /*                                                                      */
     /* parameters:                                                          */
     /*   value     ... string that contains list of arguments for set call  */
     /*   type_args ... array containing info about arguments                */
     /*   args      ... pointers to argument strings                         */
     /*   max_args  ... maximal number of argments                           */
     /*                                                                      */
     /* tokenize string with arguments and store the pointer to each         */
     /* argument in 'args'. Evaluate the type of each argument and store it  */
     /* in 'type_args'. We have the following types:                         */
     /*    't' ... a single token, or                                        */
     /*    'L' ... a list, indicated by (...)                                */
     /*                                                                      */
     /* return:                                                              */
     /*   number of parsed arguments                                         */
     /*                                                                      */
     /* error:                                                               */
     /*   return -1                                                           */
     /*----------------------------------------------------------------------*/
{
  char *token;             /* pointer to token in string */
  char *next;              /* pointer to next token in string */
  int n_args;              /* number of arguments in set call */

  /* initialize pointers */
  n_args = 0;
  *type_args = '\0';
  *args = NULL;

  /* we have two types of arguments:
     't' ... a single token, or
     'L' ... a list, indicated by (...)
  */

  /* split value string into separate arguments */
  if (value && *value != '\0') {
    /* there is at least one argument */

    for ( token = next = value;
	  next != NULL && *token != '\0' && n_args < max_args;
	  token = next, n_args++ ) {
      
      /* get type of argument */

      if ( *token == '(' ) {
	/* argument is a list */
	type_args[n_args] = 'L';
	/* skip to next character */
	token++;
	/* store pointer to argument */
	args[n_args] = token;
	/* find end of list using ')' as terminating character */
	next = strchr(token,')');
	if (next != NULL) {
	  /* end of list found */
	  *next = '\0';      /* terminate token string    */
	  token = ++next;    /* set pointer to first character after ')' */
	  /* first character after ')' must be ',' */
	  if ( !(*token == ',' || *token == '\0') ) {
	    _unur_error(GENTYPE,UNUR_ERR_STR_SYNTAX,") not followed by comma"); 
	    return -1;
	  }
	}
	else {
	  /* no ')' found */
	  token = NULL;
	}
      }

      else if ( *token == '"' ) {
	/* argument is a string */
	type_args[n_args] = 's';
	/* skip to next character */
	token++;
	/* store pointer to argument */
	args[n_args] = token;

	/* find end of string using '"' as terminating character */
	next = strchr(token,'"');
	if (next != NULL) {
	  /* end of list found */
	  *next = '\0';      /* terminate token string    */
	  token = ++next;    /* set pointer to first character after ')' */
	  /* first character after ')' must be ',' */
	  if ( !(*token == ',' || *token == '\0') ) {
	    _unur_error(GENTYPE,UNUR_ERR_STR_SYNTAX,"closing '\"' not followed by comma"); 
	    return -1;
	  }
	}
	else {
	  /* no '"' found */
	  token = NULL;
	}
      }

      else {
	/* argument is a sigle item */
	type_args[n_args] = 't';
	/* store pointer to argument */
	args[n_args] = token;
      }

      /* find end of argument string using ',' as separator */
      if (token) {
	next = strchr(token,',');
	if (next != NULL) {
	  /* next separator found */
	  *next = '\0';     /* terminate token string    */
	  next++;           /* set pointer to next token */
	}
      }
      else {
	next = NULL;
      }
    }

    /* terminate string */
    type_args[n_args] = '\0';

    /* If there are two many arguments, we simply assume a syntax error */
    /* (This case should not happen.)                                   */
    if (n_args >= max_args) { 
      _unur_error( GENTYPE, UNUR_ERR_STR_SYNTAX, "too many arguments");
      return -1;
    }
  }

  /* return number of parsed arguments */
  return n_args;

} /* end of _unur_str_set_args() */

/*---------------------------------------------------------------------------*/

/*****************************************************************************/
/**  Convert to numbers                                                     **/
/*****************************************************************************/

/*---------------------------------------------------------------------------*/

int 
_unur_parse_ilist( char *liststr, int **iarray )
     /*----------------------------------------------------------------------*/
     /* Process a comma separated list of integers, convert it to ints and   */
     /* store it an array. The list (string) is terminated by a closing      */
     /* parenthesis ')' or '\0'.                                             */
     /*                                                                      */
     /* parameters:                                                          */
     /*   liststr ... string containing list of numbers                      */
     /*   iarray  ... pointer to adress of double array for storing result   */
     /*                                                                      */
     /* return:                                                              */
     /*   number of elements extracted                                       */
     /*                                                                      */
     /* comment:                                                             */
     /*   as a side effect an int array of appropriate size is allocated.    */
     /*----------------------------------------------------------------------*/
{
  int *iarr = NULL;     /* pointer to int array */
  int n_iarray = 0;     /* number of elements in processed list */
  int n_alloc = 0;      /* number allocated elements */
  char *token;

  /* return 0 if no list is given */
  if (liststr == NULL) {
    *iarray = iarr;
    return 0;
  }

  /* discard possible leading ',' or '(' characters in front of the numbers */
  while ( *liststr == ',' || *liststr == '(' ){
    liststr++; /* next char */
  }

  /* new get integers */
  for ( token  = strtok(liststr, ",)"); 
        token != NULL;     
        token  = strtok(NULL, ",)") ) {

    /* check if there is enough space in iarr */
    if (n_iarray >= n_alloc) {
      /* no --> allocate memory */
      n_alloc += 100;
      iarr = _unur_xrealloc( iarr, n_alloc * sizeof(double) );
    }
    
    /* get int */
    iarr[n_iarray++] = _unur_atoi(token);

  } /* end while -- all elements of list read */

  /* store pointer to double array */
  *iarray = iarr;

  /* return number of elements in array */
  return n_iarray;

} /* end of _unur_parse_ilist() */

/*---------------------------------------------------------------------------*/

int 
_unur_parse_dlist( char *liststr, double **darray )
     /*----------------------------------------------------------------------*/
     /* Process a comma separated list of numbers (doubles), convert it to   */
     /* doubles and store it an array. The list (string) is terminated by    */
     /* a closing parenthesis ')' or '\0'.                                   */
     /*                                                                      */
     /* parameters:                                                          */
     /*   liststr ... string containing list of numbers                      */
     /*   darray  ... pointer to adress of double array for storing result   */
     /*                                                                      */
     /* return:                                                              */
     /*   number of elements extracted                                       */
     /*                                                                      */
     /* comment:                                                             */
     /*   as a side effect a double array of appropriate size is allocated.  */
     /*----------------------------------------------------------------------*/
{
  double *darr = NULL;  /* pointer to double array */
  int n_darray = 0;     /* number of elements in processed list */
  int n_alloc = 0;      /* number allocated elements */
  char *token;          /* pointer to token in string */
  char *next;           /* pointer to next token in string */

  /* return 0 if no list is given */
  if (liststr == NULL) {
    *darray = NULL;
    return 0;
  }

  /* extract doubles from string and write them to array of doubles.  */
  /* end of list is indicated by right bracket ')' or '\0'.           */  

  /* pointer to first token in string */
  token = liststr;

  /* remove (optional) leading parenthesis '('                        */
  while (*token != '\0' && *token == '(')
    ++token;

  /* tokenize the string using ',' as separator and handle the tokens */
  for ( next = token;
	next != NULL && *token != '\0';
	token = next ) {

    /* find end of token (i.e. the pointer to next token) */
    next = strchr(token,',');
    if (next != NULL) {
      /* next separator found */
      *next = '\0';     /* terminate token string    */
      next++;           /* set pointer to next token */
    }

   /* check if there is enough space in darr */
    if (n_darray >= n_alloc) {
      /* no --> allocate memory */
      n_alloc += 100;
      darr = _unur_xrealloc( darr, n_alloc * sizeof(double) );
    }

    /* extract double and write to array and increase counter */
    darr[n_darray] = _unur_atod(token);
    n_darray++;

  }

  /* store pointer to double array */
  *darray = darr;

  /* return number of elements in array */
  return n_darray;

} /* end of _unur_parse_dlist() */

/*---------------------------------------------------------------------------*/

int 
_unur_atoi ( const char *str )
     /*----------------------------------------------------------------------*/
     /* Converts the string into its integer representation.                 */
     /* 'true' and 'on' are converted to 1, 'false' and 'off' to 0.          */
     /*                                                                      */
     /* parameters:                                                          */
     /*   str      ... pointer to string                                     */
     /*                                                                      */
     /* return:                                                              */
     /*   integer                                                            */
     /*----------------------------------------------------------------------*/
{
  if ( !strcmp(str,"true") || !strcmp(str,"on") )
    return 1;

  else if ( !strcmp(str,"false") || !strcmp(str,"off") )
    return 0;

  else if ( !strncmp(str,"inf",3) )
    return INT_MAX;

  else if ( !strncmp(str,"-inf",4) )
    return INT_MIN;

  else
    return atoi(str);

} /* end of _unur_atoi() */

/*---------------------------------------------------------------------------*/

unsigned
_unur_atou ( const char *str )
     /*----------------------------------------------------------------------*/
     /* Converts the string into its unsigned representation.                */
     /* strings are interprated as hexadecimal numbers.                      */
     /* 'true' and 'on' are converted to 1u, 'false' and 'off' to 0u.        */
     /*                                                                      */
     /* parameters:                                                          */
     /*   str      ... pointer to string                                     */
     /*                                                                      */
     /* return:                                                              */
     /*   unsigned                                                           */
     /*----------------------------------------------------------------------*/
{
  char *tail;

  if ( !strcmp(str,"true") || !strcmp(str,"on") )
    return 1u;

  else if ( !strcmp(str,"false") || !strcmp(str,"off") )
    return 0u;

  else
    return ((unsigned) strtoul(str, &tail, 16));
} /* end of _unur_atou() */

/*---------------------------------------------------------------------------*/

double
_unur_atod ( const char *str )
     /*----------------------------------------------------------------------*/
     /* Converts the string into its double value.                           */
     /*                                                                      */
     /* parameters:                                                          */
     /*   str      ... pointer to string                                     */
     /*                                                                      */
     /* return:                                                              */
     /*   double                                                             */
     /*----------------------------------------------------------------------*/
{
  if ( !strncmp(str,"inf",3) )
    return INFINITY;

  else if ( !strncmp(str,"-inf",4) )
    return -INFINITY;

  else
    return atof(str);
} /* end of _unur_atod() */

/*---------------------------------------------------------------------------*/

/*****************************************************************************/
/**  Debugging utilities                                                    **/
/*****************************************************************************/

/*---------------------------------------------------------------------------*/
#ifdef UNUR_ENABLE_LOGGING
/*---------------------------------------------------------------------------*/

void 
_unur_str_debug_string( int level, const char *key, const char *value  )
     /*----------------------------------------------------------------------*/
     /* print key & value info into logfile                                  */
     /*                                                                      */
     /* parameters:                                                          */
     /*   level ... level of indentation                                     */
     /*   key   ... pointer to key string                                    */
     /*   value ... pointer to value string                                  */
     /*----------------------------------------------------------------------*/
{
  FILE *log;
  
  log = unur_get_stream();

/*    fprintf(log,"%s: String Interface for UNURAN\n",GENTYPE); */
  fprintf(log,"%s: ",GENTYPE);
  for (; level>0; level--) 
    fprintf(log,"\t");
  fprintf(log,"%s",key);
  if (value)
    fprintf(log,": %s",value);
  fprintf(log,"\n");

/*    fprintf(log,"%s:\n",GENTYPE); */

} /* end of _unur_str_debug_string() */

/*---------------------------------------------------------------------------*/

void 
_unur_str_debug_distr( int level, const char *name, double *params, int n_params )
     /*----------------------------------------------------------------------*/
     /* write info about distribution into logfile                           */
     /*                                                                      */
     /* parameters:                                                          */
     /*   level    ... level of indentation                                  */
     /*   name     ... name of distribution                                  */
     /*   params   ... array of parameters                                   */
     /*   n_params ... number of parameters                                  */
     /*----------------------------------------------------------------------*/
{
  FILE *log;
  int i;
  
  log = unur_get_stream();

  fprintf(log,"%s: ",GENTYPE);
  for (; level>0; level--) 
    fprintf(log,"\t");

  fprintf(log,"distribution = %s (",name);
  if (n_params >0) {
    fprintf(log,"%g",params[0]);
    for (i=1; i<n_params; i++)
      fprintf(log,", %g",params[i]);
  }
  fprintf(log,")\n");

} /* end of _unur_str_debug_distr() */

/*---------------------------------------------------------------------------*/

void
_unur_str_debug_set( int level, const char *key, const char *type, ... )
     /*----------------------------------------------------------------------*/
     /* write info about set command into logfile                            */
     /*                                                                      */
     /* parameters:                                                          */
     /*   level ... level of indentation                                     */
     /*   key   ... pointer to key string                                    */
     /*   type  ... list of types of argments                                */
     /*   ...   ... variable list of arguments                               */
     /*----------------------------------------------------------------------*/
{
  va_list ap;
  FILE *log;
  
  va_start(ap, type);

  log = unur_get_stream();

  /* print key name */
  fprintf(log,"%s: ",GENTYPE);
  for (; level>0; level--) 
    fprintf(log,"\t");
  fprintf(log,"%s: ",key);

  while (1) {

    switch (*type) {
    case 'v':
      fprintf(log," -none-");
      break;
    case 'd':
      fprintf(log," %g",va_arg(ap,double));
      break;
    case 'i':
      fprintf(log," %d",va_arg(ap,int));
      break;
    case 'u':
      fprintf(log," %x",va_arg(ap,unsigned int));
      break;
    case 'C':
      fprintf(log," %s",va_arg(ap,char *));
      break;
    case 'D': {
      int i,size;
      double *darray;
      darray = va_arg(ap,double *);
      size = va_arg(ap,int);
      if (size > 0 && darray != NULL) {
	fprintf(log," (%g",darray[0]);
	for (i=1; i<size; i++)
	  fprintf(log,",%g",darray[i]);
	fprintf(log,")");
      }
      else
	fprintf(log," (empty)");
      break;
    }
    case '!':
    default:
      fprintf(log," syntax error");
      break;
    }

    if ( *(++type) != '\0' )
      /* skip. there is a next argument to be processed */
      fprintf(log,",");
    else
      /* done */
      break;
  }

  /* dd, iD, Di */

  fprintf(log,"\n");
  fflush(log);   /* in case of a segmentation fault */

  va_end(ap);

} /* end of _unur_str_debug_set() */

/*---------------------------------------------------------------------------*/
#endif   /* end UNUR_ENABLE_LOGGING */
/*---------------------------------------------------------------------------*/
